from __future__ import annotations

import abc
from typing import Any

from moto.stepfunctions.parser.asl.component.common.variable_sample import (
    VariableSample,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.comparison_operator_type import (
    ComparisonOperatorType,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.comparison_type import (
    Comparison,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.operator.factory import (
    OperatorFactory,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.operator.operator import (
    Operator,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class ComparisonFunc(Comparison, abc.ABC):
    operator_type: ComparisonOperatorType

    def __init__(self, operator_type: ComparisonOperatorType):
        self.operator_type = operator_type


class ComparisonFuncValue(ComparisonFunc):
    value: Any

    def __init__(self, operator_type: ComparisonOperatorType, value: Any):
        super().__init__(operator_type=operator_type)
        self.value = value

    def _eval_body(self, env: Environment) -> None:
        operator: Operator = OperatorFactory.get(self.operator_type)
        operator.eval(env=env, value=self.value)


class ComparisonFuncVar(ComparisonFuncValue):
    _COMPARISON_FUNC_VAR_VALUE: str = "$"
    variable_sample: VariableSample

    def __init__(
        self, operator_type: ComparisonOperatorType, variable_sample: VariableSample
    ):
        super().__init__(
            operator_type=operator_type, value=self._COMPARISON_FUNC_VAR_VALUE
        )
        self.variable_sample = variable_sample

    def _eval_body(self, env: Environment) -> None:
        self.variable_sample.eval(env=env)
        super()._eval_body(env=env)
        # Purge the outcome of the variable sampling form the
        # stack as operators do not digest the input value.
        del env.stack[-2]



"""Handles incoming kafka requests, invokes methods, returns responses."""

import json
from urllib.parse import unquote

from moto.core.responses import BaseResponse

from .models import KafkaBackend, kafka_backends


class KafkaResponse(BaseResponse):
    """Handler for Kafka requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="kafka")

    @property
    def kafka_backend(self) -> KafkaBackend:
        """Return backend instance specific for this region."""
        return kafka_backends[self.current_account][self.region]

    def create_cluster_v2(self) -> str:
        cluster_name = self._get_param("clusterName")
        tags = self._get_param("tags")
        provisioned = self._get_param("provisioned")
        serverless = self._get_param("serverless")
        cluster_arn, cluster_name, state, cluster_type = (
            self.kafka_backend.create_cluster_v2(
                cluster_name=cluster_name,
                tags=tags,
                provisioned=provisioned,
                serverless=serverless,
            )
        )
        return json.dumps(
            dict(
                clusterArn=cluster_arn,
                clusterName=cluster_name,
                state=state,
                clusterType=cluster_type,
            )
        )

    def describe_cluster_v2(self) -> str:
        cluster_arn = unquote(self.parsed_url.path.split("/clusters/")[-1])
        cluster_info = self.kafka_backend.describe_cluster_v2(
            cluster_arn=cluster_arn,
        )
        return json.dumps(dict(clusterInfo=cluster_info))

    def list_clusters_v2(self) -> str:
        cluster_name_filter = self._get_param("clusterNameFilter")
        cluster_type_filter = self._get_param("clusterTypeFilter")
        max_results = self._get_param("maxResults")
        next_token = self._get_param("nextToken")
        cluster_info_list, next_token = self.kafka_backend.list_clusters_v2(
            cluster_name_filter=cluster_name_filter,
            cluster_type_filter=cluster_type_filter,
            max_results=max_results,
            next_token=next_token,
        )
        return json.dumps(dict(clusterInfoList=cluster_info_list, nextToken=next_token))

    def list_tags_for_resource(self) -> str:
        resource_arn = unquote(self.parsed_url.path.split("/tags/")[-1])
        tags = self.kafka_backend.list_tags_for_resource(
            resource_arn=resource_arn,
        )
        return json.dumps(dict(tags=tags))

    def tag_resource(self) -> str:
        resource_arn = unquote(self._get_param("resourceArn"))
        tags = self._get_param("tags")
        self.kafka_backend.tag_resource(
            resource_arn=resource_arn,
            tags=tags,
        )
        return json.dumps(dict())

    def untag_resource(self) -> str:
        resource_arn = unquote(self._get_param("resourceArn"))
        tag_keys = self.__dict__["data"]["tagKeys"]
        self.kafka_backend.untag_resource(
            resource_arn=resource_arn,
            tag_keys=tag_keys,
        )
        return json.dumps(dict())

    def create_cluster(self) -> str:
        broker_node_group_info = self._get_param("brokerNodeGroupInfo")
        client_authentication = self._get_param("clientAuthentication")
        cluster_name = self._get_param("clusterName")
        configuration_info = self._get_param("configurationInfo")
        encryption_info = self._get_param("encryptionInfo")
        enhanced_monitoring = self._get_param("enhancedMonitoring")
        open_monitoring = self._get_param("openMonitoring")
        kafka_version = self._get_param("kafkaVersion")
        logging_info = self._get_param("loggingInfo")
        number_of_broker_nodes = self._get_param("numberOfBrokerNodes")
        tags = self._get_param("tags")
        storage_mode = self._get_param("storageMode")
        cluster_arn, cluster_name, state = self.kafka_backend.create_cluster(
            broker_node_group_info=broker_node_group_info,
            client_authentication=client_authentication,
            cluster_name=cluster_name,
            configuration_info=configuration_info,
            encryption_info=encryption_info,
            enhanced_monitoring=enhanced_monitoring,
            open_monitoring=open_monitoring,
            kafka_version=kafka_version,
            logging_info=logging_info,
            number_of_broker_nodes=number_of_broker_nodes,
            tags=tags,
            storage_mode=storage_mode,
        )
        return json.dumps(
            dict(clusterArn=cluster_arn, clusterName=cluster_name, state=state)
        )

    def describe_cluster(self) -> str:
        cluster_arn = unquote(self.parsed_url.path.split("/clusters/")[-1])
        cluster_info = self.kafka_backend.describe_cluster(
            cluster_arn=cluster_arn,
        )
        return json.dumps(dict(clusterInfo=cluster_info))

    def delete_cluster(self) -> str:
        cluster_arn = unquote(self.parsed_url.path.split("/clusters/")[-1])
        current_version = self._get_param("currentVersion")
        cluster_arn, state = self.kafka_backend.delete_cluster(
            cluster_arn=cluster_arn,
            current_version=current_version,
        )
        return json.dumps(dict(clusterArn=cluster_arn, state=state))

    def list_clusters(self) -> str:
        cluster_name_filter = self._get_param("clusterNameFilter")
        max_results = self._get_param("maxResults")
        next_token = self._get_param("nextToken")

        cluster_info_list = self.kafka_backend.list_clusters(
            cluster_name_filter=cluster_name_filter,
            max_results=max_results,
            next_token=next_token,
        )

        return json.dumps(dict(clusterInfoList=cluster_info_list, nextToken=next_token))



import json

import moto.server as server


def test_create_without_enable_option():
    backend = server.create_backend_app("guardduty")
    test_client = backend.test_client()

    body = {"enable": "True"}
    response = test_client.post("/detector", data=json.dumps(body))
    assert response.status_code == 200
    assert "detectorId" in json.loads(response.data)



"""AppMeshBackend class with methods for supported APIs."""

from typing import Dict, List, Literal, Optional, Union

from moto.appmesh.dataclasses.mesh import (
    Mesh,
    MeshSpec,
)
from moto.appmesh.dataclasses.route import Route, RouteMetadata, RouteSpec
from moto.appmesh.dataclasses.shared import Metadata
from moto.appmesh.dataclasses.virtual_node import (
    VirtualNode,
    VirtualNodeMetadata,
    VirtualNodeSpec,
)
from moto.appmesh.dataclasses.virtual_router import (
    PortMapping,
    VirtualRouter,
    VirtualRouterSpec,
)
from moto.appmesh.exceptions import (
    MeshNotFoundError,
    MeshOwnerDoesNotMatchError,
    ResourceNotFoundError,
    RouteNameAlreadyTakenError,
    RouteNotFoundError,
    VirtualNodeNameAlreadyTakenError,
    VirtualNodeNotFoundError,
    VirtualRouterNameAlreadyTakenError,
    VirtualRouterNotFoundError,
)
from moto.core.base_backend import BackendDict, BaseBackend
from moto.utilities.paginator import paginate

PAGINATION_MODEL = {
    "list_meshes": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 100,
        "unique_attribute": "meshName",
    },
    "list_tags_for_resource": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 100,
        "unique_attribute": ["key", "value"],
    },
    "list_virtual_routers": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 100,
        "unique_attribute": "virtualRouterName",
    },
    "list_routes": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 100,
        "unique_attribute": "route_name",
    },
    "list_virtual_nodes": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 100,
        "unique_attribute": "virtual_node_name",
    },
}


class AppMeshBackend(BaseBackend):
    """Implementation of AppMesh APIs."""

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.meshes: Dict[str, Mesh] = dict()

    def _validate_mesh(self, mesh_name: str, mesh_owner: Optional[str]) -> None:
        if mesh_name not in self.meshes:
            raise MeshNotFoundError(mesh_name=mesh_name)
        if (
            mesh_owner is not None
            and self.meshes[mesh_name].metadata.mesh_owner != mesh_owner
        ):
            raise MeshOwnerDoesNotMatchError(mesh_name, mesh_owner)

    def _check_virtual_node_validity(
        self,
        mesh_name: str,
        mesh_owner: Optional[str],
        virtual_node_name: str,
    ) -> None:
        self._validate_mesh(mesh_name=mesh_name, mesh_owner=mesh_owner)
        if virtual_node_name not in self.meshes[mesh_name].virtual_nodes:
            raise VirtualNodeNotFoundError(mesh_name, virtual_node_name)
        return

    def _check_virtual_node_availability(
        self,
        mesh_name: str,
        mesh_owner: Optional[str],
        virtual_node_name: str,
    ) -> None:
        self._validate_mesh(mesh_name=mesh_name, mesh_owner=mesh_owner)
        if virtual_node_name in self.meshes[mesh_name].virtual_nodes:
            raise VirtualNodeNameAlreadyTakenError(
                mesh_name=mesh_name, virtual_node_name=virtual_node_name
            )
        return

    def _check_router_availability(
        self,
        mesh_name: str,
        mesh_owner: Optional[str],
        virtual_router_name: str,
    ) -> None:
        self._validate_mesh(mesh_name=mesh_name, mesh_owner=mesh_owner)
        if virtual_router_name in self.meshes[mesh_name].virtual_routers:
            raise VirtualRouterNameAlreadyTakenError(
                virtual_router_name=virtual_router_name, mesh_name=mesh_name
            )
        return

    def _check_router_validity(
        self,
        mesh_name: str,
        mesh_owner: Optional[str],
        virtual_router_name: str,
    ) -> None:
        self._validate_mesh(mesh_name=mesh_name, mesh_owner=mesh_owner)
        if virtual_router_name not in self.meshes[mesh_name].virtual_routers:
            raise VirtualRouterNotFoundError(
                virtual_router_name=virtual_router_name, mesh_name=mesh_name
            )
        return

    def _check_route_validity(
        self,
        mesh_name: str,
        mesh_owner: Optional[str],
        virtual_router_name: str,
        route_name: str,
    ) -> None:
        self._check_router_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_router_name=virtual_router_name,
        )
        if (
            route_name
            not in self.meshes[mesh_name].virtual_routers[virtual_router_name].routes
        ):
            raise RouteNotFoundError(
                mesh_name=mesh_name,
                virtual_router_name=virtual_router_name,
                route_name=route_name,
            )
        return

    def _check_route_availability(
        self,
        mesh_name: str,
        mesh_owner: Optional[str],
        virtual_router_name: str,
        route_name: str,
    ) -> None:
        self._check_router_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_router_name=virtual_router_name,
        )
        if (
            route_name
            in self.meshes[mesh_name].virtual_routers[virtual_router_name].routes
        ):
            raise RouteNameAlreadyTakenError(
                mesh_name=mesh_name,
                virtual_router_name=virtual_router_name,
                route_name=route_name,
            )
        return

    def create_mesh(
        self,
        client_token: Optional[str],
        mesh_name: str,
        egress_filter_type: Optional[str],
        ip_preference: Optional[str],
        tags: Optional[List[Dict[str, str]]],
    ) -> Mesh:
        from moto.sts import sts_backends

        sts_backend = sts_backends[self.account_id]["global"]
        user_id, _, _ = sts_backend.get_caller_identity(
            self.account_id, region=self.region_name
        )

        metadata = Metadata(
            arn=f"arn:aws:appmesh:{self.region_name}:{self.account_id}:{mesh_name}",
            mesh_owner=user_id,
            resource_owner=user_id,
        )
        spec = MeshSpec(
            egress_filter={"type": egress_filter_type},
            service_discovery={"ip_preference": ip_preference},
        )
        mesh = Mesh(
            mesh_name=mesh_name,
            spec=spec,
            status={"status": "ACTIVE"},
            metadata=metadata,
            tags=tags or list(),
        )
        self.meshes[mesh_name] = mesh
        return mesh

    def update_mesh(
        self,
        client_token: Optional[str],
        mesh_name: str,
        egress_filter_type: Optional[str],
        ip_preference: Optional[str],
    ) -> Mesh:
        if mesh_name not in self.meshes:
            raise MeshNotFoundError(mesh_name=mesh_name)
        updated = False
        if egress_filter_type is not None:
            self.meshes[mesh_name].spec.egress_filter["type"] = egress_filter_type
            updated = True

        if ip_preference is not None:
            self.meshes[mesh_name].spec.service_discovery["ip_preference"] = (
                ip_preference
            )
            updated = True

        if updated:
            self.meshes[mesh_name].metadata.update_timestamp()
            self.meshes[mesh_name].metadata.version += 1
        return self.meshes[mesh_name]

    def describe_mesh(self, mesh_name: str, mesh_owner: Optional[str]) -> Mesh:
        self._validate_mesh(mesh_name=mesh_name, mesh_owner=mesh_owner)
        return self.meshes[mesh_name]

    def delete_mesh(self, mesh_name: str) -> Mesh:
        if mesh_name not in self.meshes:
            raise MeshNotFoundError(mesh_name=mesh_name)
        self.meshes[mesh_name].status["status"] = "DELETED"
        mesh = self.meshes[mesh_name]
        del self.meshes[mesh_name]
        return mesh

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_meshes(self) -> List[Dict[str, Union[str, int]]]:
        return [
            {
                "arn": mesh.metadata.arn,
                "createdAt": mesh.metadata.created_at.strftime("%d/%m/%Y, %H:%M:%S"),
                "lastUpdatedAt": mesh.metadata.last_updated_at.strftime(
                    "%d/%m/%Y, %H:%M:%S"
                ),
                "meshName": mesh.mesh_name,
                "meshOwner": mesh.metadata.mesh_owner,
                "resourceOwner": mesh.metadata.resource_owner,
                "version": mesh.metadata.version,
            }
            for mesh in self.meshes.values()
        ]

    def _get_resource_with_arn(
        self, resource_arn: str
    ) -> Union[Mesh, VirtualRouter, Route, VirtualNode]:
        for mesh in self.meshes.values():
            if mesh.metadata.arn == resource_arn:
                return mesh
            for virtual_router in mesh.virtual_routers.values():
                if virtual_router.metadata.arn == resource_arn:
                    return virtual_router
                for route in virtual_router.routes.values():
                    if route.metadata.arn == resource_arn:
                        return route
            for virtual_node in mesh.virtual_nodes.values():
                if virtual_node.metadata.arn == resource_arn:
                    return virtual_node
        raise ResourceNotFoundError(resource_arn)

    @paginate(pagination_model=PAGINATION_MODEL)  # type: ignore
    def list_tags_for_resource(self, resource_arn: str) -> List[Dict[str, str]]:
        return self._get_resource_with_arn(resource_arn=resource_arn).tags

    def tag_resource(self, resource_arn: str, tags: List[Dict[str, str]]) -> None:
        if len(tags) > 0:
            resource = self._get_resource_with_arn(resource_arn=resource_arn)
            resource.tags.extend(tags)
        return

    def describe_virtual_router(
        self, mesh_name: str, mesh_owner: Optional[str], virtual_router_name: str
    ) -> VirtualRouter:
        self._check_router_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_router_name=virtual_router_name,
        )
        return self.meshes[mesh_name].virtual_routers[virtual_router_name]

    def create_virtual_router(
        self,
        client_token: str,
        mesh_name: str,
        mesh_owner: Optional[str],
        port_mappings: List[PortMapping],
        tags: Optional[List[Dict[str, str]]],
        virtual_router_name: str,
    ) -> VirtualRouter:
        self._check_router_availability(
            mesh_name=mesh_name,
            virtual_router_name=virtual_router_name,
            mesh_owner=mesh_owner,
        )
        owner = mesh_owner or self.meshes[mesh_name].metadata.mesh_owner
        metadata = Metadata(
            mesh_owner=owner,
            resource_owner=owner,
            arn=f"arn:aws:appmesh:{self.region_name}:{self.account_id}:mesh/{mesh_name}/virtualRouter/{virtual_router_name}",
        )
        listeners: List[Dict[Literal["port_mapping"], PortMapping]] = [
            {"port_mapping": port_mapping} for port_mapping in port_mappings
        ]
        spec = VirtualRouterSpec(listeners=listeners)
        virtual_router = VirtualRouter(
            virtual_router_name=virtual_router_name,
            mesh_name=mesh_name,
            metadata=metadata,
            status={"status": "ACTIVE"},
            spec=spec,
            tags=tags or list(),
        )
        self.meshes[mesh_name].virtual_routers[virtual_router_name] = virtual_router
        return virtual_router

    def update_virtual_router(
        self,
        client_token: str,
        mesh_name: str,
        mesh_owner: Optional[str],
        port_mappings: List[PortMapping],
        virtual_router_name: str,
    ) -> VirtualRouter:
        self._check_router_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_router_name=virtual_router_name,
        )
        listeners: List[Dict[Literal["port_mapping"], PortMapping]] = [
            {"port_mapping": port_mapping} for port_mapping in port_mappings
        ]
        spec = VirtualRouterSpec(listeners=listeners)
        virtual_router = self.meshes[mesh_name].virtual_routers[virtual_router_name]
        virtual_router.spec = spec
        virtual_router.metadata.update_timestamp()
        virtual_router.metadata.version += 1
        return virtual_router

    def delete_virtual_router(
        self, mesh_name: str, mesh_owner: Optional[str], virtual_router_name: str
    ) -> VirtualRouter:
        self._check_router_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_router_name=virtual_router_name,
        )
        mesh = self.meshes[mesh_name]
        mesh.virtual_routers[virtual_router_name].status["status"] = "DELETED"
        virtual_router = mesh.virtual_routers[virtual_router_name]
        del mesh.virtual_routers[virtual_router_name]
        return virtual_router

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_virtual_routers(
        self, mesh_name: str, mesh_owner: Optional[str]
    ) -> List[Dict[str, Union[str, int]]]:
        self._validate_mesh(mesh_name=mesh_name, mesh_owner=mesh_owner)
        return [
            {
                "arn": virtual_router.metadata.arn,
                "createdAt": virtual_router.metadata.created_at.strftime(
                    "%d/%m/%Y, %H:%M:%S"
                ),
                "lastUpdatedAt": virtual_router.metadata.last_updated_at.strftime(
                    "%d/%m/%Y, %H:%M:%S"
                ),
                "meshName": virtual_router.mesh_name,
                "meshOwner": virtual_router.metadata.mesh_owner,
                "resourceOwner": virtual_router.metadata.resource_owner,
                "version": virtual_router.metadata.version,
                "virtualRouterName": virtual_router.virtual_router_name,
            }
            for virtual_router in self.meshes[mesh_name].virtual_routers.values()
        ]

    def create_route(
        self,
        client_token: Optional[str],
        mesh_name: str,
        mesh_owner: Optional[str],
        route_name: str,
        spec: RouteSpec,
        tags: Optional[List[Dict[str, str]]],
        virtual_router_name: str,
    ) -> Route:
        self._check_route_availability(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            route_name=route_name,
            virtual_router_name=virtual_router_name,
        )
        owner = mesh_owner or self.meshes[mesh_name].metadata.mesh_owner
        metadata = RouteMetadata(
            arn=f"arn:aws:appmesh:{self.region_name}:{self.account_id}:mesh/{mesh_name}/virtualRouter/{virtual_router_name}/route/{route_name}",
            mesh_name=mesh_name,
            mesh_owner=owner,
            resource_owner=owner,
            route_name=route_name,
            virtual_router_name=virtual_router_name,
        )
        route = Route(
            mesh_name=mesh_name,
            mesh_owner=owner,
            metadata=metadata,
            route_name=route_name,
            spec=spec,
            tags=tags or list(),
            virtual_router_name=virtual_router_name,
        )
        self.meshes[mesh_name].virtual_routers[virtual_router_name].routes[
            route_name
        ] = route
        return route

    def describe_route(
        self,
        mesh_name: str,
        mesh_owner: Optional[str],
        route_name: str,
        virtual_router_name: str,
    ) -> Route:
        self._check_route_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_router_name=virtual_router_name,
            route_name=route_name,
        )
        return (
            self.meshes[mesh_name]
            .virtual_routers[virtual_router_name]
            .routes[route_name]
        )

    def update_route(
        self,
        client_token: Optional[str],
        mesh_name: str,
        mesh_owner: Optional[str],
        route_name: str,
        spec: RouteSpec,
        virtual_router_name: str,
    ) -> Route:
        self._check_route_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_router_name=virtual_router_name,
            route_name=route_name,
        )
        route = (
            self.meshes[mesh_name]
            .virtual_routers[virtual_router_name]
            .routes[route_name]
        )
        route.spec = spec
        route.metadata.version += 1
        route.metadata.update_timestamp()
        return route

    def delete_route(
        self,
        mesh_name: str,
        mesh_owner: Optional[str],
        route_name: str,
        virtual_router_name: str,
    ) -> Route:
        self._check_route_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_router_name=virtual_router_name,
            route_name=route_name,
        )
        route = (
            self.meshes[mesh_name]
            .virtual_routers[virtual_router_name]
            .routes[route_name]
        )
        route.status["status"] = "DELETED"
        del (
            self.meshes[mesh_name]
            .virtual_routers[virtual_router_name]
            .routes[route_name]
        )
        return route

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_routes(
        self,
        mesh_name: str,
        mesh_owner: Optional[str],
        virtual_router_name: str,
    ) -> List[RouteMetadata]:
        self._check_router_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_router_name=virtual_router_name,
        )
        virtual_router = self.meshes[mesh_name].virtual_routers[virtual_router_name]
        return [route.metadata for route in virtual_router.routes.values()]

    def describe_virtual_node(
        self, mesh_name: str, mesh_owner: Optional[str], virtual_node_name: str
    ) -> VirtualNode:
        self._check_virtual_node_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_node_name=virtual_node_name,
        )
        return self.meshes[mesh_name].virtual_nodes[virtual_node_name]

    def create_virtual_node(
        self,
        client_token: Optional[str],
        mesh_name: str,
        mesh_owner: Optional[str],
        spec: VirtualNodeSpec,
        tags: Optional[List[Dict[str, str]]],
        virtual_node_name: str,
    ) -> VirtualNode:
        self._check_virtual_node_availability(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_node_name=virtual_node_name,
        )
        owner = mesh_owner or self.meshes[mesh_name].metadata.mesh_owner
        metadata = VirtualNodeMetadata(
            arn=f"arn:aws:appmesh:{self.region_name}:{self.account_id}:mesh/{mesh_name}/virtualNode/{virtual_node_name}",
            mesh_name=mesh_name,
            mesh_owner=owner,
            resource_owner=owner,
            virtual_node_name=virtual_node_name,
        )
        virtual_node = VirtualNode(
            mesh_name=mesh_name,
            mesh_owner=owner,
            metadata=metadata,
            spec=spec,
            tags=tags or list(),
            virtual_node_name=virtual_node_name,
        )
        self.meshes[mesh_name].virtual_nodes[virtual_node_name] = virtual_node
        return virtual_node

    def update_virtual_node(
        self,
        client_token: Optional[str],
        mesh_name: str,
        mesh_owner: Optional[str],
        spec: VirtualNodeSpec,
        virtual_node_name: str,
    ) -> VirtualNode:
        self._check_virtual_node_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_node_name=virtual_node_name,
        )
        virtual_node = self.meshes[mesh_name].virtual_nodes[virtual_node_name]
        virtual_node.spec = spec
        virtual_node.metadata.version += 1
        virtual_node.metadata.update_timestamp()
        return virtual_node

    def delete_virtual_node(
        self, mesh_name: str, mesh_owner: Optional[str], virtual_node_name: str
    ) -> VirtualNode:
        self._check_virtual_node_validity(
            mesh_name=mesh_name,
            mesh_owner=mesh_owner,
            virtual_node_name=virtual_node_name,
        )
        virtual_node = self.meshes[mesh_name].virtual_nodes[virtual_node_name]
        virtual_node.status["status"] = "DELETED"
        del self.meshes[mesh_name].virtual_nodes[virtual_node_name]
        return virtual_node

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_virtual_nodes(
        self,
        mesh_name: str,
        mesh_owner: Optional[str],
    ) -> List[VirtualNodeMetadata]:
        self._validate_mesh(mesh_name=mesh_name, mesh_owner=mesh_owner)
        virtual_nodes = self.meshes[mesh_name].virtual_nodes
        return [virtual_node.metadata for virtual_node in virtual_nodes.values()]


appmesh_backends = BackendDict(AppMeshBackend, "appmesh")



import os
from unittest import SkipTest, mock

import boto3

from moto import mock_aws, settings
from tests import EXAMPLE_AMI_PARAVIRTUAL, EXAMPLE_AMI_WINDOWS


# The default AMIs are not loaded for our test case, to speed things up
# But we do need it for this specific test (and others in this file..)
@mock.patch.dict(os.environ, {"MOTO_EC2_LOAD_DEFAULT_AMIS": "true"})
@mock_aws
def test_get_password_data():
    if settings.TEST_SERVER_MODE:
        raise SkipTest("Can't set environment variables in ServerMode")
    client = boto3.client("ec2", region_name="us-east-1")

    # Ensure non-windows instances return empty password data
    instance_id = client.run_instances(
        ImageId=EXAMPLE_AMI_PARAVIRTUAL, MinCount=1, MaxCount=1
    )["Instances"][0]["InstanceId"]
    resp = client.get_password_data(InstanceId=instance_id)
    assert resp["InstanceId"] == instance_id
    assert resp["PasswordData"] == ""

    # Ensure Windows instances
    instance_id = client.run_instances(
        ImageId=EXAMPLE_AMI_WINDOWS, MinCount=1, MaxCount=1
    )["Instances"][0]["InstanceId"]
    resp = client.get_password_data(InstanceId=instance_id)
    assert resp["InstanceId"] == instance_id
    assert len(resp["PasswordData"]) == 128



import json
from typing import List
from unittest import SkipTest
from uuid import uuid4

import boto3
import pytest

from moto import mock_aws, settings
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID
from moto.s3.models import FakeBucket, FakeKey
from moto.s3.notifications import (
    S3NotificationEvent,
    _detail_type,
    _send_event_bridge_message,
)

REGION_NAME = "us-east-1"


@pytest.mark.parametrize(
    "event_names, expected_event_message",
    [
        (
            [
                S3NotificationEvent.OBJECT_CREATED_PUT_EVENT,
                S3NotificationEvent.OBJECT_CREATED_POST_EVENT,
                S3NotificationEvent.OBJECT_CREATED_COPY_EVENT,
                S3NotificationEvent.OBJECT_CREATED_COMPLETE_MULTIPART_UPLOAD_EVENT,
            ],
            "Object Created",
        ),
        (
            [
                S3NotificationEvent.OBJECT_REMOVED_DELETE_EVENT,
                S3NotificationEvent.OBJECT_REMOVED_DELETE_MARKER_CREATED_EVENT,
            ],
            "Object Deleted",
        ),
        ([S3NotificationEvent.OBJECT_RESTORE_POST_EVENT], "Object Restore Initiated"),
        (
            [S3NotificationEvent.OBJECT_RESTORE_COMPLETED_EVENT],
            "Object Restore Completed",
        ),
        (
            [S3NotificationEvent.OBJECT_RESTORE_DELETE_EVENT],
            "Object Restore Expired",
        ),
        (
            [S3NotificationEvent.LIFECYCLE_TRANSITION_EVENT],
            "Object Storage Class Changed",
        ),
        ([S3NotificationEvent.INTELLIGENT_TIERING_EVENT], "Object Access Tier Changed"),
        ([S3NotificationEvent.OBJECT_ACL_UPDATE_EVENT], "Object ACL Updated"),
        ([S3NotificationEvent.OBJECT_TAGGING_PUT_EVENT], "Object Tags Added"),
        ([S3NotificationEvent.OBJECT_TAGGING_DELETE_EVENT], "Object Tags Deleted"),
    ],
)
def test_detail_type(event_names: List[str], expected_event_message: str):
    for event_name in event_names:
        assert _detail_type(event_name) == expected_event_message


def test_detail_type_unknown_event():
    with pytest.raises(ValueError) as ex:
        _detail_type("unknown event")
    assert (
        str(ex.value)
        == "unsupported event `unknown event` for s3 eventbridge notification (https://docs.aws.amazon.com/AmazonS3/latest/userguide/EventBridge.html)"
    )


@mock_aws
def test_send_event_bridge_message():
    # setup mocks
    events_client = boto3.client("events", region_name=REGION_NAME)
    logs_client = boto3.client("logs", region_name=REGION_NAME)
    rule_name = "test-rule"
    events_client.put_rule(
        Name=rule_name, EventPattern=json.dumps({"account": [ACCOUNT_ID]})
    )
    log_group_name = "/test-group"
    logs_client.create_log_group(logGroupName=log_group_name)
    mocked_bucket = FakeBucket(str(uuid4()), ACCOUNT_ID, REGION_NAME)
    mocked_key = FakeKey(
        "test-key",
        bytes("test content", encoding="utf-8"),
        ACCOUNT_ID,
        region_name=REGION_NAME,
    )

    # do nothing if event target does not exists.
    _send_event_bridge_message(
        ACCOUNT_ID,
        mocked_bucket,
        S3NotificationEvent.OBJECT_CREATED_PUT_EVENT,
        mocked_key,
    )
    assert (
        len(logs_client.filter_log_events(logGroupName=log_group_name)["events"]) == 0
    )

    # do nothing even if an error is raised while sending events.
    events_client.put_targets(
        Rule=rule_name,
        Targets=[
            {
                "Id": "test",
                "Arn": f"arn:aws:logs:{REGION_NAME}:{ACCOUNT_ID}:log-group:{log_group_name}",
            }
        ],
    )

    _send_event_bridge_message(ACCOUNT_ID, mocked_bucket, "unknown-event", mocked_key)
    assert (
        len(logs_client.filter_log_events(logGroupName=log_group_name)["events"]) == 0
    )

    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest(("Doesn't quite work right with the Proxy or Server"))
    # an event is correctly sent to the log group.
    _send_event_bridge_message(
        ACCOUNT_ID,
        mocked_bucket,
        S3NotificationEvent.OBJECT_CREATED_PUT_EVENT,
        mocked_key,
    )
    events = logs_client.filter_log_events(logGroupName=log_group_name)["events"]
    assert len(events) == 1
    event_msg = json.loads(events[0]["message"])
    assert event_msg["detail-type"] == "Object Created"
    assert event_msg["source"] == "aws.s3"
    assert event_msg["region"] == REGION_NAME
    assert event_msg["resources"] == [f"arn:aws:s3:::{mocked_bucket.name}"]
    event_detail = event_msg["detail"]
    assert event_detail["bucket"] == {"name": mocked_bucket.name}
    assert event_detail["object"]["key"] == mocked_key.name
    assert event_detail["reason"] == "ObjectCreated"



from __future__ import annotations

from typing import Final

from moto.stepfunctions.parser.asl.component.component import Component
from moto.stepfunctions.parser.asl.component.state.exec.state_map.execution_type import (
    ExecutionType,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.mode import (
    Mode,
)


class ProcessorConfig(Component):
    DEFAULT_MODE: Final[Mode] = Mode.Inline
    DEFAULT_EXECUTION_TYPE: Final[ExecutionType] = ExecutionType.Standard

    mode: Final[Mode]
    execution_type: Final[ExecutionType]

    def __init__(
        self,
        mode: Mode = DEFAULT_MODE,
        execution_type: ExecutionType = DEFAULT_EXECUTION_TYPE,
    ):
        super().__init__()
        self.mode = mode
        self.execution_type = execution_type



from typing import Set

from moto.stepfunctions.parser.asl.component.common.query_language import (
    QueryLanguageMode,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.jsonata.jsonata import (
    JSONataExpression,
    VariableDeclarations,
    VariableReference,
    compose_jsonata_expression,
    eval_jsonata_expression,
    extract_jsonata_variable_references,
)


class VariableSample(EvalComponent):
    query_language_mode: QueryLanguageMode
    expression: str

    def __init__(self, query_language_mode: QueryLanguageMode, expression: str):
        super().__init__()
        # TODO: check for illegal functions ($, $$, $eval)
        self.query_language_mode = query_language_mode
        self.expression = expression

    def _eval_body(self, env: Environment) -> None:
        # Get the variables sampled in the jsonata expression.
        expression_variable_references: Set[VariableReference] = (
            extract_jsonata_variable_references(self.expression)
        )
        variable_declarations_list = list()
        if self.query_language_mode == QueryLanguageMode.JSONata:
            # Sample $states values into expression.
            states_variable_declarations: VariableDeclarations = (
                env.states.to_variable_declarations(
                    variable_references=expression_variable_references
                )
            )
            variable_declarations_list.append(states_variable_declarations)

        # Sample Variable store values in to expression.
        # TODO: this could be optimised by sampling only those invoked.
        variable_declarations: VariableDeclarations = (
            env.variable_store.get_variable_declarations()
        )
        variable_declarations_list.append(variable_declarations)

        rich_jsonata_expression: JSONataExpression = compose_jsonata_expression(
            final_jsonata_expression=self.expression,
            variable_declarations_list=variable_declarations_list,
        )
        result = eval_jsonata_expression(rich_jsonata_expression)
        env.stack.append(result)



import json

import boto3
import pytest

from moto import mock_aws


@mock_aws
def test_create_job_template():
    client = boto3.client("iot", region_name="eu-west-1")
    job_template_id = "TestJobTemplate"

    job_document = {"field": "value"}

    job_template = client.create_job_template(
        jobTemplateId=job_template_id,
        document=json.dumps(job_document),
        description="Description",
        presignedUrlConfig={
            "roleArn": "arn:aws:iam::1:role/service-role/iot_job_role",
            "expiresInSec": 123,
        },
        jobExecutionsRolloutConfig={"maximumPerMinute": 10},
        jobExecutionsRetryConfig={
            "criteriaList": [{"failureType": "ALL", "numberOfRetries": 10}]
        },
        abortConfig={
            "criteriaList": [
                {
                    "action": "CANCEL",
                    "failureType": "ALL",
                    "minNumberOfExecutedThings": 1,
                    "thresholdPercentage": 90,
                }
            ]
        },
    )

    assert job_template["jobTemplateId"] == job_template_id
    assert (
        job_template["jobTemplateArn"]
        == "arn:aws:iot:eu-west-1:123456789012:jobtemplate/TestJobTemplate"
    )


@mock_aws
def test_create_job_template_with_invalid_id():
    client = boto3.client("iot", region_name="eu-west-1")
    job_template_id = "TestJobTemplate!@#4"

    with pytest.raises(client.exceptions.InvalidRequestException):
        client.create_job_template(
            jobTemplateId=job_template_id,
            document=json.dumps({"field": "value"}),
            description="Description",
        )


@mock_aws
def test_create_job_template_with_the_same_id_twice():
    client = boto3.client("iot", region_name="eu-west-1")
    job_template_id = "TestJobTemplate"

    client.create_job_template(
        jobTemplateId=job_template_id,
        document=json.dumps({"field": "value"}),
        description="Description",
    )

    with pytest.raises(client.exceptions.ConflictException):
        client.create_job_template(
            jobTemplateId=job_template_id,
            document=json.dumps({"field": "value"}),
            description="Description",
        )


@mock_aws
def test_describe_job_template():
    client = boto3.client("iot", region_name="eu-west-1")
    job_template_id = "TestJobTemplate"

    job_document = {"field": "value"}

    client.create_job_template(
        jobTemplateId=job_template_id,
        document=json.dumps(job_document),
        description="Description",
        presignedUrlConfig={
            "roleArn": "arn:aws:iam::1:role/service-role/iot_job_role",
            "expiresInSec": 123,
        },
        jobExecutionsRolloutConfig={"maximumPerMinute": 10},
        jobExecutionsRetryConfig={
            "criteriaList": [{"failureType": "ALL", "numberOfRetries": 10}]
        },
        abortConfig={
            "criteriaList": [
                {
                    "action": "CANCEL",
                    "failureType": "ALL",
                    "minNumberOfExecutedThings": 1,
                    "thresholdPercentage": 90,
                }
            ]
        },
    )

    job_template = client.describe_job_template(jobTemplateId=job_template_id)

    assert job_template["jobTemplateId"] == job_template_id
    assert (
        job_template["jobTemplateArn"]
        == "arn:aws:iot:eu-west-1:123456789012:jobtemplate/TestJobTemplate"
    )
    assert job_template["description"] == "Description"
    assert job_template["presignedUrlConfig"] == {
        "roleArn": "arn:aws:iam::1:role/service-role/iot_job_role",
        "expiresInSec": 123,
    }
    assert job_template["jobExecutionsRolloutConfig"] == {"maximumPerMinute": 10}
    assert job_template["jobExecutionsRetryConfig"]["criteriaList"] == [
        {"failureType": "ALL", "numberOfRetries": 10}
    ]


@mock_aws
def test_describe_nonexistent_job_template():
    client = boto3.client("iot", region_name="eu-west-1")
    with pytest.raises(client.exceptions.ResourceNotFoundException):
        client.describe_job_template(jobTemplateId="nonexistent")


@mock_aws
def test_delete_nonexistent_job_template():
    client = boto3.client("iot", region_name="eu-west-1")
    with pytest.raises(client.exceptions.ResourceNotFoundException):
        client.delete_job_template(jobTemplateId="nonexistent")


@mock_aws
def test_list_job_templates():
    client = boto3.client("iot", region_name="eu-west-1")
    job_document = {"field": "value"}
    job_template_ids = ["TestJobTemplate1", "AnotherJobTemplate"]

    for template_id in job_template_ids:
        job_template = client.create_job_template(
            jobTemplateId=template_id,
            document=json.dumps(job_document),
            description="Description",
        )

        assert job_template["jobTemplateId"] == template_id

    list_result = client.list_job_templates(maxResults=100)
    assert "nextToken" not in list_result
    assert len(list_result["jobTemplates"]) == 2


@mock_aws
def test_list_job_templates_wht_pagination():
    client = boto3.client("iot", region_name="eu-west-1")
    job_document = {"field": "value"}
    job_template_ids = [
        "TestJobTemplate1",
        "AnotherJobTemplate",
        "YetAnotherJob",
        "LasttestJob",
    ]

    for template_id in job_template_ids:
        job_template = client.create_job_template(
            jobTemplateId=template_id,
            document=json.dumps(job_document),
            description="Description",
        )

        assert job_template["jobTemplateId"] == template_id

    first_list_result = client.list_job_templates(maxResults=2)
    assert "nextToken" in first_list_result
    assert len(first_list_result["jobTemplates"]) == 2

    second_list_result = client.list_job_templates(
        maxResults=2, nextToken=first_list_result["nextToken"]
    )
    assert "nextToken" not in second_list_result
    assert len(second_list_result["jobTemplates"]) == 2


@mock_aws
def test_delete_job_template():
    # given
    client = boto3.client("iot", region_name="eu-west-1")
    job_template_id = "TestJobTemplate"
    client.create_job_template(
        jobTemplateId=job_template_id,
        document=json.dumps({"field": "value"}),
        description="Description",
    )
    list_result = client.list_job_templates(maxResults=100)
    assert len(list_result["jobTemplates"]) == 1

    # when
    client.delete_job_template(
        jobTemplateId=job_template_id,
    )

    # then
    list_result = client.list_job_templates(maxResults=100)
    assert len(list_result["jobTemplates"]) == 0



import json

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws


def get_policy_doc(action, stream_arn):
    sts = boto3.client("sts", "us-east-1")
    account_id = sts.get_caller_identity()["Account"]

    policy_doc = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Principal": {"AWS": account_id},
                "Effect": "Allow",
                "Action": [action],
                "Resource": f"{stream_arn}",
            }
        ],
    }
    return json.dumps(policy_doc)


@mock_aws
def test_put_resource_policy():
    client = boto3.client("kinesis", region_name="us-west-2")
    client.create_stream(StreamName="my-stream", ShardCount=2)
    resp = client.describe_stream(StreamName="my-stream")
    desc = resp["StreamDescription"]
    stream_arn = desc["StreamARN"]

    action = "kinesis:*"
    json_policy_doc = get_policy_doc(action, stream_arn)
    client.put_resource_policy(ResourceARN=stream_arn, Policy=json_policy_doc)
    response = client.get_resource_policy(ResourceARN=stream_arn)

    assert response["Policy"] == json_policy_doc


@mock_aws
def test_delete_resource_policy():
    client = boto3.client("kinesis", region_name="us-west-2")
    client.create_stream(StreamName="my-stream", ShardCount=2)
    resp = client.describe_stream(StreamName="my-stream")
    desc = resp["StreamDescription"]
    stream_arn = desc["StreamARN"]
    action = "kinesis:*"
    json_policy_doc = get_policy_doc(action, stream_arn)
    client.put_resource_policy(ResourceARN=stream_arn, Policy=json_policy_doc)
    response = client.get_resource_policy(ResourceARN=stream_arn)
    assert response["Policy"] == json_policy_doc

    client.delete_resource_policy(ResourceARN=stream_arn)
    response = client.get_resource_policy(ResourceARN=stream_arn)
    assert response["Policy"] == "{}"


@mock_aws
def test_get_resource_policy_from_unknown_resource():
    client = boto3.client("kinesis", region_name="us-west-2")
    client.create_stream(StreamName="my-stream", ShardCount=2)
    resp = client.describe_stream(StreamName="my-stream")
    desc = resp["StreamDescription"]
    stream_arn = desc["StreamARN"] + "unknown"
    sts = boto3.client("sts", "us-east-1")
    account_id = sts.get_caller_identity()["Account"]
    with pytest.raises(ClientError) as exc:
        client.get_resource_policy(ResourceARN=stream_arn)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert (
        err["Message"] == f"Stream {stream_arn} under account {account_id} not found."
    )


@mock_aws
def test_delete_resource_policy_from_unknown_resource():
    client = boto3.client("kinesis", region_name="us-west-2")
    client.create_stream(StreamName="my-stream", ShardCount=2)
    resp = client.describe_stream(StreamName="my-stream")
    desc = resp["StreamDescription"]
    stream_arn = desc["StreamARN"] + "unknown"
    sts = boto3.client("sts", "us-east-1")
    account_id = sts.get_caller_identity()["Account"]
    with pytest.raises(ClientError) as exc:
        client.delete_resource_policy(ResourceARN=stream_arn)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert (
        err["Message"] == f"Stream {stream_arn} under account {account_id} not found."
    )


@mock_aws
def test_delete_resource_policy_from_resource_without_policy():
    client = boto3.client("kinesis", region_name="us-west-2")
    client.create_stream(StreamName="my-stream", ShardCount=2)
    resp = client.describe_stream(StreamName="my-stream")
    desc = resp["StreamDescription"]
    stream_arn = desc["StreamARN"]
    with pytest.raises(ClientError) as exc:
        client.delete_resource_policy(ResourceARN=stream_arn)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == f"No resource policy found for resource ARN {stream_arn}."



from threading import Timer as ThreadingTimer

from moto.core.common_models import BaseModel


class Timer(BaseModel):
    def __init__(self, background_timer: ThreadingTimer, started_event_id: int):
        self.background_timer = background_timer
        self.started_event_id = started_event_id

    def start(self) -> None:
        return self.background_timer.start()

    def is_alive(self) -> bool:
        return self.background_timer.is_alive()

    def cancel(self) -> None:
        return self.background_timer.cancel()



from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

from moto.utilities.utils import get_partition

from .exceptions import ValidationError


def make_arn_for_compute_env(account_id: str, name: str, region_name: str) -> str:
    return f"arn:{get_partition(region_name)}:batch:{region_name}:{account_id}:compute-environment/{name}"


def make_arn_for_job_queue(account_id: str, name: str, region_name: str) -> str:
    return f"arn:{get_partition(region_name)}:batch:{region_name}:{account_id}:job-queue/{name}"


def make_arn_for_job(account_id: str, job_id: str, region_name: str) -> str:
    return f"arn:{get_partition(region_name)}:batch:{region_name}:{account_id}:job/{job_id}"


def make_arn_for_task_def(
    account_id: str, name: str, revision: int, region_name: str
) -> str:
    return f"arn:{get_partition(region_name)}:batch:{region_name}:{account_id}:job-definition/{name}:{revision}"


def lowercase_first_key(some_dict: Dict[str, Any]) -> Dict[str, Any]:
    new_dict: Dict[str, Any] = {}
    for key, value in some_dict.items():
        new_key = key[0].lower() + key[1:]
        try:
            if isinstance(value, dict):
                new_dict[new_key] = lowercase_first_key(value)
            elif all([isinstance(v, dict) for v in value]):
                new_dict[new_key] = [lowercase_first_key(v) for v in value]
            else:
                new_dict[new_key] = value
        except TypeError:
            new_dict[new_key] = value

    return new_dict


def validate_job_status(target_job_status: str, valid_job_statuses: List[str]) -> None:
    if target_job_status not in valid_job_statuses:
        raise ValidationError(
            (
                "1 validation error detected: Value at 'current_status' failed "
                "to satisfy constraint: Member must satisfy enum value set: {valid_statues}"
            ).format(valid_statues=valid_job_statuses)
        )


class JobStatus(str, Enum):
    SUBMITTED = "SUBMITTED"
    PENDING = "PENDING"
    RUNNABLE = "RUNNABLE"
    STARTING = "STARTING"
    RUNNING = "RUNNING"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"

    @classmethod
    def job_statuses(self) -> List[str]:
        return sorted([item.value for item in JobStatus])

    @classmethod
    def is_job_already_started(self, current_status: str) -> bool:
        validate_job_status(current_status, JobStatus.job_statuses())
        return current_status not in [
            JobStatus.SUBMITTED,
            JobStatus.PENDING,
            JobStatus.RUNNABLE,
            JobStatus.STARTING,
        ]

    @classmethod
    def is_job_before_starting(self, current_status: str) -> bool:
        validate_job_status(current_status, JobStatus.job_statuses())
        return current_status in [
            JobStatus.SUBMITTED,
            JobStatus.PENDING,
            JobStatus.RUNNABLE,
        ]

    @classmethod
    def status_transitions(self) -> List[Tuple[Optional[str], str]]:
        return [
            (JobStatus.SUBMITTED.value, JobStatus.PENDING.value),
            (JobStatus.PENDING.value, JobStatus.RUNNABLE.value),
            (JobStatus.RUNNABLE.value, JobStatus.STARTING),
            (JobStatus.STARTING.value, JobStatus.RUNNING.value),
        ]



import json
from io import BytesIO
from typing import Any, Dict, List
from unittest import SkipTest
from uuid import uuid4

import boto3
import pytest
import requests

from moto import mock_aws, settings
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID
from moto.settings import S3_UPLOAD_PART_MIN_SIZE
from tests.test_s3.test_s3_multipart import reduced_min_part_size

REGION_NAME = "us-east-1"

if settings.TEST_DECORATOR_MODE:
    REDUCED_PART_SIZE = 256
else:
    REDUCED_PART_SIZE = S3_UPLOAD_PART_MIN_SIZE


def _seteup_bucket_notification_eventbridge(
    bucket_name: str = str(uuid4()),
    rule_name: str = "test-rule",
    log_group_name: str = "/test-group",
    region: str = REGION_NAME,
) -> Dict[str, str]:
    """Setups S3, EventBridge and CloudWatchLogs"""
    # Setup S3
    s3_res = boto3.resource("s3", region_name=region)
    if region == "us-east-1":
        s3_res.create_bucket(Bucket=bucket_name)
    else:
        s3_res.create_bucket(
            Bucket=bucket_name, CreateBucketConfiguration={"LocationConstraint": region}
        )

    # Put bucket notification event bridge
    s3_client = boto3.client("s3", region_name=region)
    s3_client.put_bucket_notification_configuration(
        Bucket=bucket_name,
        NotificationConfiguration={"EventBridgeConfiguration": {}},
    )

    # Setup EventBridge Rule
    events_client = boto3.client("events", region_name=region)
    events_client.put_rule(
        Name=rule_name, EventPattern=json.dumps({"account": [ACCOUNT_ID]})
    )

    # Create a log group and attach it to the events target.
    logs_client = boto3.client("logs", region_name=region)
    logs_client.create_log_group(logGroupName=log_group_name)
    events_client.put_targets(
        Rule=rule_name,
        Targets=[
            {
                "Id": "test",
                "Arn": f"arn:aws:logs:{region}:{ACCOUNT_ID}:log-group:{log_group_name}",
            }
        ],
    )

    return {
        "bucket_name": bucket_name,
        "event_rule_name": rule_name,
        "log_group_name": log_group_name,
    }


def _get_send_events(
    log_group_name: str = "/test-group", region: str = REGION_NAME
) -> List[Dict[str, Any]]:
    logs_client = boto3.client("logs", region_name=region)
    return sorted(
        logs_client.filter_log_events(logGroupName=log_group_name)["events"],
        key=lambda item: item["timestamp"],
    )


@mock_aws
@pytest.mark.parametrize(
    "region,partition", [("us-west-2", "aws"), ("cn-north-1", "aws-cn")]
)
def test_put_object_notification_ObjectCreated_PUT(region, partition):
    resource_names = _seteup_bucket_notification_eventbridge(region=region)
    bucket_name = resource_names["bucket_name"]
    s3_client = boto3.client("s3", region_name=region)

    # Put Object
    s3_client.put_object(Bucket=bucket_name, Key="keyname", Body="bodyofnewobject")

    events = _get_send_events(region=region)
    assert len(events) == 2
    event_message = json.loads(events[0]["message"])
    assert event_message["detail-type"] == "Object Created"
    assert event_message["source"] == "aws.s3"
    assert event_message["account"] == ACCOUNT_ID
    assert event_message["region"] == region
    assert event_message["resources"] == [f"arn:{partition}:s3:::{bucket_name}"]
    assert event_message["detail"]["bucket"]["name"] == bucket_name
    assert event_message["detail"]["reason"] == "ObjectCreated"


@mock_aws
@reduced_min_part_size
def test_put_object_notification_ObjectCreated_POST():
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest(("Doesn't quite work right with the Proxy or Server"))

    resource_names = _seteup_bucket_notification_eventbridge()
    bucket_name = resource_names["bucket_name"]

    ###
    # multipart/formdata POST request (this request is processed in S3Response._bucket_response_post)
    ###
    content = b"Hello, this is a sample content for the multipart upload."
    object_key = "test-key"

    _ = requests.post(
        f"https://{bucket_name}.s3.amazonaws.com/",
        data={"key": object_key},
        files={"file": ("tmp.txt", BytesIO(content))},
    )

    events = _get_send_events()
    assert len(events) == 1
    event_message = json.loads(events[0]["message"])
    assert event_message["detail-type"] == "Object Created"
    assert event_message["source"] == "aws.s3"
    assert event_message["account"] == ACCOUNT_ID
    assert event_message["region"] == REGION_NAME
    assert event_message["detail"]["bucket"]["name"] == bucket_name
    assert event_message["detail"]["object"]["key"] == object_key
    assert event_message["detail"]["reason"] == "ObjectCreated"


@mock_aws
def test_copy_object_notification():
    resource_names = _seteup_bucket_notification_eventbridge()
    bucket_name = resource_names["bucket_name"]
    s3_client = boto3.client("s3", region_name=REGION_NAME)

    # Copy object (send two events; PutObject and CopyObject)
    s3_client.put_object(Bucket=bucket_name, Key="keyname", Body="bodyofnewobject")
    object_key = "key2"
    s3_client.copy_object(
        Bucket=bucket_name, CopySource=f"{bucket_name}/keyname", Key="key2"
    )

    events = _get_send_events()
    assert len(events) == 3  # [PutObject, ObjectTagging, CopyObject]
    event_message = json.loads(events[-1]["message"])
    assert event_message["detail-type"] == "Object Created"
    assert event_message["source"] == "aws.s3"
    assert event_message["account"] == ACCOUNT_ID
    assert event_message["region"] == REGION_NAME
    assert event_message["detail"]["bucket"]["name"] == bucket_name
    assert event_message["detail"]["object"]["key"] == object_key
    assert event_message["detail"]["reason"] == "ObjectCreated"


@mock_aws
@reduced_min_part_size
def test_complete_multipart_upload_notification():
    resource_names = _seteup_bucket_notification_eventbridge()
    bucket_name = resource_names["bucket_name"]
    s3_client = boto3.client("s3", region_name=REGION_NAME)
    object_key = "testkey"

    part1 = b"0" * REDUCED_PART_SIZE
    part2 = b"1"
    multipart = s3_client.create_multipart_upload(Bucket=bucket_name, Key=object_key)
    up1 = s3_client.upload_part(
        Body=BytesIO(part1),
        PartNumber=1,
        Bucket=bucket_name,
        Key=object_key,
        UploadId=multipart["UploadId"],
    )
    up2 = s3_client.upload_part(
        Body=BytesIO(part2),
        PartNumber=2,
        Bucket=bucket_name,
        Key=object_key,
        UploadId=multipart["UploadId"],
    )

    s3_client.complete_multipart_upload(
        Bucket=bucket_name,
        Key=object_key,
        MultipartUpload={
            "Parts": [
                {"ETag": up1["ETag"], "PartNumber": 1},
                {"ETag": up2["ETag"], "PartNumber": 2},
            ]
        },
        UploadId=multipart["UploadId"],
    )

    events = _get_send_events()
    assert len(events) == 3  # [PutObject, ObjectTagging, CompleteMultipartUpload]
    event_message = json.loads(events[-1]["message"])
    assert event_message["detail-type"] == "Object Created"
    assert event_message["source"] == "aws.s3"
    assert event_message["account"] == ACCOUNT_ID
    assert event_message["region"] == REGION_NAME
    assert event_message["detail"]["bucket"]["name"] == bucket_name
    assert event_message["detail"]["object"]["key"] == object_key
    assert event_message["detail"]["reason"] == "ObjectCreated"


@mock_aws
def test_delete_object_notification():
    resource_names = _seteup_bucket_notification_eventbridge()
    bucket_name = resource_names["bucket_name"]
    s3_client = boto3.client("s3", region_name=REGION_NAME)

    # Put Object
    s3_client.put_object(Bucket=bucket_name, Key="keyname", Body="bodyofnewobject")

    # Delete Object
    s3_client.delete_object(Bucket=bucket_name, Key="keyname")

    events = _get_send_events()
    assert len(events) == 3  # [PutObject, ObjectTagging, DeleteObject]
    event_message = json.loads(events[-1]["message"])
    assert event_message["detail-type"] == "Object Deleted"
    assert event_message["source"] == "aws.s3"
    assert event_message["account"] == ACCOUNT_ID
    assert event_message["region"] == REGION_NAME
    assert event_message["detail"]["bucket"]["name"] == bucket_name
    assert event_message["detail"]["reason"] == "ObjectRemoved"


@mock_aws
def test_restore_key_notifications():
    resource_names = _seteup_bucket_notification_eventbridge()
    bucket_name = resource_names["bucket_name"]

    s3_resource = boto3.resource("s3", region_name=REGION_NAME)

    bucket = s3_resource.Bucket(bucket_name)
    key = bucket.put_object(Key="the-key", Body=b"somedata", StorageClass="GLACIER")
    key.restore_object(RestoreRequest={"Days": 1})

    events = _get_send_events()
    event_names = [json.loads(e["message"])["detail"]["reason"] for e in events]
    assert event_names == ["ObjectCreated", "ObjectTagging", "ObjectRestore"]

    # Finish the Object Restoration - restore Completes immediately by default
    key.load()

    events = _get_send_events()
    event_names = [json.loads(e["message"])["detail"]["reason"] for e in events]
    assert event_names == [
        "ObjectCreated",
        "ObjectTagging",
        "ObjectRestore",
        "ObjectRestore",
    ]

    # Sanity check - loading the Key does not mean the Restore-event is fired every time
    key.load()

    events = _get_send_events()
    event_names = [json.loads(e["message"])["detail"]["reason"] for e in events]
    assert event_names == [
        "ObjectCreated",
        "ObjectTagging",
        "ObjectRestore",
        "ObjectRestore",
    ]


@mock_aws
def test_put_object_tagging_notification():
    resource_names = _seteup_bucket_notification_eventbridge()
    bucket_name = resource_names["bucket_name"]
    s3_client = boto3.client("s3", region_name=REGION_NAME)

    # Put Object
    s3_client.put_object(Bucket=bucket_name, Key="keyname", Body="bodyofnewobject")

    # Put Object Tagging
    s3_client.put_object_tagging(
        Bucket=bucket_name,
        Key="keyname",
        Tagging={
            "TagSet": [
                {"Key": "item1", "Value": "foo"},
                {"Key": "item2", "Value": "bar"},
            ]
        },
    )

    events = _get_send_events()
    assert len(events) == 3
    event_message = json.loads(events[2]["message"])
    assert event_message["detail-type"] == "Object Tags Added"
    assert event_message["source"] == "aws.s3"
    assert event_message["account"] == ACCOUNT_ID
    assert event_message["region"] == REGION_NAME
    assert event_message["detail"]["bucket"]["name"] == bucket_name
    assert event_message["detail"]["reason"] == "ObjectTagging"


@mock_aws
def test_delete_object_tagging_notification():
    resource_names = _seteup_bucket_notification_eventbridge()
    bucket_name = resource_names["bucket_name"]
    s3_client = boto3.client("s3", region_name=REGION_NAME)

    # Put Object
    s3_client.put_object(Bucket=bucket_name, Key="keyname", Body="bodyofnewobject")

    # Put Object Tagging
    s3_client.put_object_tagging(
        Bucket=bucket_name,
        Key="keyname",
        Tagging={
            "TagSet": [
                {"Key": "item1", "Value": "foo"},
                {"Key": "item2", "Value": "bar"},
            ]
        },
    )

    # Delete Object Tagging
    s3_client.delete_object_tagging(Bucket=bucket_name, Key="keyname")

    events = _get_send_events()
    assert len(events) == 4
    event_message = json.loads(events[3]["message"])
    assert event_message["detail-type"] == "Object Tags Deleted"
    assert event_message["source"] == "aws.s3"
    assert event_message["account"] == ACCOUNT_ID
    assert event_message["region"] == REGION_NAME
    assert event_message["detail"]["bucket"]["name"] == bucket_name
    assert event_message["detail"]["reason"] == "ObjectTagging"


@mock_aws
def test_delete_object_acl_update_notification():
    resource_names = _seteup_bucket_notification_eventbridge()
    bucket_name = resource_names["bucket_name"]
    s3_client = boto3.client("s3", region_name=REGION_NAME)

    # Put Object
    s3_client.put_object(Bucket=bucket_name, Key="keyname", Body="bodyofnewobject")

    # Put Object ACL
    s3_client.put_object_acl(Bucket=bucket_name, Key="keyname", ACL="public-read")

    # Delete Object ACL
    s3_client.put_object_acl(Bucket=bucket_name, Key="keyname", ACL="private")

    events = _get_send_events()
    assert (
        len(events) == 4
    )  # [PutObject, ObjectTagging, CompleteMultipartUpload, ObjectAcl]
    event_message = json.loads(events[2]["message"])
    assert event_message["detail-type"] == "Object ACL Updated"
    assert event_message["source"] == "aws.s3"
    assert event_message["account"] == ACCOUNT_ID
    assert event_message["region"] == REGION_NAME
    assert event_message["detail"]["bucket"]["name"] == bucket_name
    assert event_message["detail"]["object"]["key"] == "keyname"
    assert event_message["detail"]["reason"] == "ObjectAcl"



from typing import Final, List

from moto.stepfunctions.parser.asl.component.common.error_name.error_name import (
    ErrorName,
)
from moto.stepfunctions.parser.asl.component.common.retry.retrier_decl import (
    RetrierDecl,
)
from moto.stepfunctions.parser.asl.component.common.retry.retrier_outcome import (
    RetrierOutcome,
)
from moto.stepfunctions.parser.asl.component.common.retry.retry_outcome import (
    RetryOutcome,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class RetryDecl(EvalComponent):
    def __init__(self, retriers: List[RetrierDecl]):
        self.retriers: Final[List[RetrierDecl]] = retriers

    def _eval_body(self, env: Environment) -> None:
        error_name: ErrorName = env.stack.pop()

        for retrier in self.retriers:
            env.stack.append(error_name)
            retrier.eval(env)
            outcome: RetrierOutcome = env.stack.pop()

            if outcome == RetrierOutcome.Skipped:
                continue
            elif outcome == RetrierOutcome.Executed:
                env.stack.append(RetryOutcome.CanRetry)
                return
            elif outcome == RetrierOutcome.Failed:
                env.stack.append(RetryOutcome.CannotRetry)
                return

        env.stack.append(RetryOutcome.NoRetrier)



import boto3

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_list_signing_platforms():
    client = boto3.client("signer", region_name="us-east-2")
    resp = client.list_signing_platforms()

    assert "platforms" in resp
    assert len(resp["platforms"]) == 4

    partners = [x["partner"] for x in resp["platforms"]]
    assert set(partners) == {"AmazonFreeRTOS", "AWSLambda", "AWSIoTDeviceManagement"}



"""Test the different server responses."""

import json
import re
import unittest

import pytest
import xmltodict

import moto.server as server
from moto import mock_aws, settings


@pytest.fixture(scope="function", autouse=True)
def skip_in_server_mode():
    if settings.TEST_SERVER_MODE:
        raise unittest.SkipTest("No point in testing this in ServerMode")


@mock_aws
def test_describe_clusters():
    backend = server.create_backend_app("redshift")
    test_client = backend.test_client()

    res = test_client.get("/?Action=DescribeClusters")

    result = res.data.decode("utf-8")
    assert "<Clusters></Clusters>" in result


@mock_aws
def test_describe_clusters_with_json_content_type():
    backend = server.create_backend_app("redshift")
    test_client = backend.test_client()

    res = test_client.get("/?Action=DescribeClusters&ContentType=JSON")

    result = json.loads(res.data.decode("utf-8"))
    del result["DescribeClustersResponse"]["ResponseMetadata"]
    assert result == {
        "DescribeClustersResponse": {"DescribeClustersResult": {"Clusters": []}}
    }


@pytest.mark.parametrize("is_json", [True, False], ids=["JSON", "XML"])
@mock_aws
def test_create_cluster(is_json):
    backend = server.create_backend_app("redshift")
    test_client = backend.test_client()

    create_params = (
        "?Action=CreateCluster"
        "&ClusterIdentifier=examplecluster"
        "&MasterUsername=masteruser"
        "&MasterUserPassword=12345678Aa"
        "&NodeType=ds2.xlarge"
    )
    if is_json:
        create_params += "&ContentType=JSON"
    res = test_client.post(create_params)

    result = res.data.decode("utf-8")

    if is_json:
        result = json.loads(result)
    else:
        result = xmltodict.parse(result, dict_constructor=dict)

    del result["CreateClusterResponse"]["ResponseMetadata"]
    assert "CreateClusterResponse" in result
    assert "CreateClusterResult" in result["CreateClusterResponse"]
    assert "Cluster" in result["CreateClusterResponse"]["CreateClusterResult"]
    result = result["CreateClusterResponse"]["CreateClusterResult"]["Cluster"]

    assert result["MasterUsername"] == "masteruser"
    assert result["MasterUserPassword"] == "****"
    assert result["ClusterVersion"] == "1.0"
    assert result["ClusterSubnetGroupName"] is None
    assert result["AvailabilityZone"] == "us-east-1a"
    assert result["ClusterStatus"] == "creating"
    assert result["NumberOfNodes"] == (1 if is_json else "1")
    assert result["PubliclyAccessible"] is None
    assert result["Encrypted"] is None
    assert result["DBName"] == "dev"
    assert result["NodeType"] == "ds2.xlarge"
    assert result["ClusterIdentifier"] == "examplecluster"
    assert re.match(
        "examplecluster.[a-z0-9]+.us-east-1.redshift.amazonaws.com",
        result["Endpoint"]["Address"],
    )
    assert result["Endpoint"]["Port"] == (5439 if is_json else "5439")
    assert "ClusterCreateTime" in result


@pytest.mark.parametrize("is_json", [True, False], ids=["JSON", "XML"])
@mock_aws
def test_create_cluster_multiple_params(is_json):
    backend = server.create_backend_app("redshift")
    test_client = backend.test_client()

    create_params = (
        "?Action=CreateCluster"
        "&ClusterIdentifier=examplecluster"
        "&MasterUsername=masteruser"
        "&MasterUserPassword=12345678Aa"
        "&NumberOfNodes=3"
        "&NodeType=ds2.xlarge"
        "&EnhancedVpcRouting=True"
        "&Tags.Tag.1.Key=key1"
        "&Tags.Tag.1.Value=val1"
        "&Tags.Tag.2.Key=key2"
        "&Tags.Tag.2.Value=val2"
        "&DBName=testdb"
        "&Encrypted=True"
        "&ClusterVersion=2.0"
        "&Port=1234"
    )
    if is_json:
        create_params += "&ContentType=JSON"
    res = test_client.post(create_params)

    result = res.data.decode("utf-8")

    if is_json:
        result = json.loads(result)
    else:
        result = xmltodict.parse(result, dict_constructor=dict)

    del result["CreateClusterResponse"]["ResponseMetadata"]
    assert "CreateClusterResponse" in result
    assert "CreateClusterResult" in result["CreateClusterResponse"]
    assert "Cluster" in result["CreateClusterResponse"]["CreateClusterResult"]
    result = result["CreateClusterResponse"]["CreateClusterResult"]["Cluster"]

    assert result["MasterUsername"] == "masteruser"
    assert result["MasterUserPassword"] == "****"
    assert result["ClusterVersion"] == "2.0"
    assert result["ClusterSubnetGroupName"] is None
    assert result["AvailabilityZone"] == "us-east-1a"
    assert result["ClusterStatus"] == "creating"
    assert result["NumberOfNodes"] == (3 if is_json else "3")
    assert result["PubliclyAccessible"] is None
    assert result["Encrypted"] == "True"
    assert result["DBName"] == "testdb"
    assert result["NodeType"] == "ds2.xlarge"
    assert result["ClusterIdentifier"] == "examplecluster"
    assert re.match(
        "examplecluster.[a-z0-9]+.us-east-1.redshift.amazonaws.com",
        result["Endpoint"]["Address"],
    )
    assert result["Endpoint"]["Port"] == (1234 if is_json else "1234")
    assert "ClusterCreateTime" in result
    assert "Tags" in result
    tags = result["Tags"]
    if not is_json:
        tags = tags["item"]
    assert tags == [{"Key": "key1", "Value": "val1"}, {"Key": "key2", "Value": "val2"}]


@pytest.mark.parametrize("is_json", [True, False], ids=["JSON", "XML"])
@mock_aws
def test_create_and_describe_clusters(is_json):
    backend = server.create_backend_app("redshift")
    test_client = backend.test_client()
    cluster_names = ["examplecluster1", "examplecluster2"]

    for name in cluster_names:
        create_params = (
            "?Action=CreateCluster"
            "&ClusterIdentifier=" + name + "&MasterUsername=masteruser"
            "&MasterUserPassword=12345678Aa"
            "&NodeType=ds2.xlarge"
        )
        if is_json:
            create_params += "&ContentType=JSON"
        test_client.post(create_params)

    describe_params = "/?Action=DescribeClusters"
    if is_json:
        describe_params += "&ContentType=JSON"
    res = test_client.get(describe_params)
    result = res.data.decode("utf-8")

    if is_json:
        result = json.loads(result)
    else:
        result = xmltodict.parse(result, dict_constructor=dict)

    del result["DescribeClustersResponse"]["ResponseMetadata"]
    assert "DescribeClustersResponse" in result
    assert "DescribeClustersResult" in result["DescribeClustersResponse"]
    assert "Clusters" in result["DescribeClustersResponse"]["DescribeClustersResult"]
    result = result["DescribeClustersResponse"]["DescribeClustersResult"]["Clusters"]
    if not is_json:
        result = result["item"]

    assert len(result) == 2
    for cluster in result:
        assert cluster["ClusterIdentifier"] in cluster_names


@pytest.mark.parametrize("is_json", [True, False], ids=["JSON", "XML"])
@mock_aws
def test_create_and_describe_cluster_security_group(is_json):
    backend = server.create_backend_app("redshift")
    test_client = backend.test_client()
    security_group_names = ["csg1", "csg2"]

    for csg in security_group_names:
        create_params = (
            "?Action=CreateClusterSecurityGroup"
            "&ClusterSecurityGroupName=" + csg + "&Description=desc for " + csg + ""
        )
        if is_json:
            create_params += "&ContentType=JSON"
        test_client.post(create_params)

    describe_params = "/?Action=DescribeClusterSecurityGroups"
    if is_json:
        describe_params += "&ContentType=JSON"
    res = test_client.get(describe_params)
    result = res.data.decode("utf-8")

    if is_json:
        result = json.loads(result)
    else:
        result = xmltodict.parse(result, dict_constructor=dict)

    groups = result["DescribeClusterSecurityGroupsResponse"][
        "DescribeClusterSecurityGroupsResult"
    ]["ClusterSecurityGroups"]
    if not is_json:
        groups = groups["item"]

    descriptions = [g["Description"] for g in groups]
    assert "desc for csg1" in descriptions
    assert "desc for csg2" in descriptions

    # Describe single SG
    describe_params = (
        "/?Action=DescribeClusterSecurityGroups&ClusterSecurityGroupName=csg1"
    )
    if is_json:
        describe_params += "&ContentType=JSON"
    res = test_client.get(describe_params)
    result = res.data.decode("utf-8")

    if is_json:
        result = json.loads(result)
    else:
        result = xmltodict.parse(result, dict_constructor=dict)

    groups = result["DescribeClusterSecurityGroupsResponse"][
        "DescribeClusterSecurityGroupsResult"
    ]["ClusterSecurityGroups"]

    if is_json:
        assert len(groups) == 1
        assert groups[0]["ClusterSecurityGroupName"] == "csg1"
    else:
        assert groups["item"]["ClusterSecurityGroupName"] == "csg1"


@pytest.mark.parametrize("is_json", [True, False], ids=["JSON", "XML"])
@mock_aws
def test_describe_unknown_cluster_security_group(is_json):
    backend = server.create_backend_app("redshift")
    test_client = backend.test_client()

    describe_params = (
        "/?Action=DescribeClusterSecurityGroups&ClusterSecurityGroupName=unknown"
    )
    if is_json:
        describe_params += "&ContentType=JSON"
    res = test_client.get(describe_params)

    assert res.status_code == 400

    if is_json:
        response = json.loads(res.data.decode("utf-8"))
    else:
        response = xmltodict.parse(res.data.decode("utf-8"), dict_constructor=dict)[
            "RedshiftClientError"
        ]
    error = response["Error"]

    assert error["Code"] == "ClusterSecurityGroupNotFound"
    assert error["Message"] == "Security group unknown not found."


@pytest.mark.parametrize("is_json", [True, False], ids=["JSON", "XML"])
@mock_aws
def test_create_cluster_with_security_group(is_json):
    backend = server.create_backend_app("redshift")
    test_client = backend.test_client()
    security_group_names = ["csg1", "csg2"]

    for csg in security_group_names:
        create_params = (
            "?Action=CreateClusterSecurityGroup"
            "&ClusterSecurityGroupName=" + csg + "&Description=desc for " + csg + ""
        )
        if is_json:
            create_params += "&ContentType=JSON"
        res = test_client.post(create_params)

        response = res.data.decode("utf-8")

        if is_json:
            response = json.loads(response)
        else:
            response = xmltodict.parse(response, dict_constructor=dict)

        del response["CreateClusterSecurityGroupResponse"]["ResponseMetadata"]
        assert "CreateClusterSecurityGroupResponse" in response
        response = response["CreateClusterSecurityGroupResponse"]
        assert "CreateClusterSecurityGroupResult" in response
        result = response["CreateClusterSecurityGroupResult"]
        assert "ClusterSecurityGroup" in result
        sg = result["ClusterSecurityGroup"]
        assert sg["ClusterSecurityGroupName"] == csg
        assert sg["Description"] == "desc for " + csg
        assert sg["EC2SecurityGroups"] == ([] if is_json else None)

    # Create Cluster with these security groups
    create_params = (
        "?Action=CreateCluster"
        "&ClusterIdentifier=examplecluster"
        "&MasterUsername=masteruser"
        "&MasterUserPassword=12345678Aa"
        "&NodeType=ds2.xlarge"
        "&ClusterSecurityGroups.member.1=csg1"
        "&ClusterSecurityGroups.member.2=csg2"
    )
    if is_json:
        create_params += "&ContentType=JSON"
    res = test_client.post(create_params)

    result = res.data.decode("utf-8")

    if is_json:
        result = json.loads(result)
    else:
        result = xmltodict.parse(result, dict_constructor=dict)

    del result["CreateClusterResponse"]["ResponseMetadata"]
    assert "CreateClusterResponse" in result
    assert "CreateClusterResult" in result["CreateClusterResponse"]
    assert "Cluster" in result["CreateClusterResponse"]["CreateClusterResult"]
    result = result["CreateClusterResponse"]["CreateClusterResult"]["Cluster"]

    security_groups = result["ClusterSecurityGroups"]
    if not is_json:
        security_groups = security_groups["item"]

    assert len(security_groups) == 2
    for csg in security_group_names:
        assert {"ClusterSecurityGroupName": csg, "Status": "active"} in security_groups



import boto3
import pytest

from moto import mock_aws


@pytest.fixture(name="client")
def client_fixture():
    with mock_aws():
        yield boto3.client("iot", region_name="eu-west-1")


def test_create_role_alias(client):
    role_alias_name = "test-role-alias"
    create_response = client.create_role_alias(
        roleAlias=role_alias_name,
        roleArn="arn:aws:iam::123456789012:role/my-role",
        credentialDurationSeconds=1234,
    )

    assert create_response["roleAlias"] == role_alias_name
    assert (
        create_response["roleAliasArn"]
        == f"arn:aws:iot:eu-west-1:123456789012:rolealias/{role_alias_name}"
    )

    assert len(client.list_role_aliases()["roleAliases"]) == 1


def test_create_role_alias_twice(client):
    role_alias_name = "test-role-alias"
    create_response = client.create_role_alias(
        roleAlias=role_alias_name,
        roleArn="arn:aws:iam::123456789012:role/my-role",
        credentialDurationSeconds=1234,
    )

    assert create_response["roleAlias"] == role_alias_name
    assert (
        create_response["roleAliasArn"]
        == f"arn:aws:iot:eu-west-1:123456789012:rolealias/{role_alias_name}"
    )

    with pytest.raises(client.exceptions.ResourceAlreadyExistsException):
        client.create_role_alias(
            roleAlias=role_alias_name,
            roleArn="arn:aws:iam::123456789012:role/my-role",
            credentialDurationSeconds=1234,
        )


def test_list_role_aliases(client):
    client.create_role_alias(
        roleAlias="test-role-alias", roleArn="arn:aws:iam::123456789012:role/my-role"
    )
    client.create_role_alias(
        roleAlias="another_role_alias",
        roleArn="arn:aws:iam::123456789012:role/my-role",
    )

    response = client.list_role_aliases()

    assert response["roleAliases"] == ["test-role-alias", "another_role_alias"]


def test_delete_role_alias(client):
    role_alias_name = "test-role-alias"

    client.create_role_alias(
        roleAlias=role_alias_name, roleArn="arn:aws:iam::123456789012:role/my-role"
    )
    assert len(client.list_role_aliases()["roleAliases"]) == 1

    client.delete_role_alias(roleAlias=role_alias_name)
    assert len(client.list_role_aliases()["roleAliases"]) == 0


def test_delete_nonexistent_role_alias(client):
    with pytest.raises(client.exceptions.ResourceNotFoundException):
        client.delete_role_alias(roleAlias="test_role_alias")


def test_describe_role_alias(client):
    role_alias_name = "test-role-alias"
    client.create_role_alias(
        roleAlias=role_alias_name, roleArn="arn:aws:iam::123456789012:role/my-role"
    )

    response = client.describe_role_alias(roleAlias=role_alias_name)
    assert response["roleAliasDescription"]["roleAlias"] == role_alias_name
    assert (
        response["roleAliasDescription"]["roleAliasArn"]
        == f"arn:aws:iot:eu-west-1:123456789012:rolealias/{role_alias_name}"
    )
    assert (
        response["roleAliasDescription"]["roleArn"]
        == "arn:aws:iam::123456789012:role/my-role"
    )
    assert response["roleAliasDescription"]["credentialDurationSeconds"] == 3600
    assert "owner" in response["roleAliasDescription"]
    assert "creationDate" in response["roleAliasDescription"]
    assert "lastModifiedDate" in response["roleAliasDescription"]


def test_update_role_alias(client):
    role_alias_name = "test-role-alias"
    client.create_role_alias(
        roleAlias=role_alias_name,
        roleArn="arn:aws:iam::123456789012:role/my-role",
        credentialDurationSeconds=1234,
    )
    client.update_role_alias(
        roleAlias=role_alias_name,
        roleArn="arn:aws:iam::123456789012:role/other-role",
        credentialDurationSeconds=2345,
    )
    response = client.describe_role_alias(roleAlias=role_alias_name)

    assert response["roleAliasDescription"]["roleAlias"] == role_alias_name
    assert (
        response["roleAliasDescription"]["roleAliasArn"]
        == f"arn:aws:iot:eu-west-1:123456789012:rolealias/{role_alias_name}"
    )
    assert (
        response["roleAliasDescription"]["roleArn"]
        == "arn:aws:iam::123456789012:role/other-role"
    )
    assert response["roleAliasDescription"]["credentialDurationSeconds"] == 2345



from typing import Any

from moto.stepfunctions.parser.asl.component.state.choice.comparison.comparison_operator_type import (
    ComparisonOperatorType,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.operator.implementations.is_operator import (
    IsTimestamp,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.operator.operator import (
    Operator,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.utils.json_path import extract_json


class TimestampEquals(Operator):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.TimestampEquals)

    @staticmethod
    def _compare(variable: Any, comparison_value: Any) -> bool:
        res = False
        if isinstance(variable, str):
            a = IsTimestamp.string_to_timestamp(variable)
            if a is not None:
                b = IsTimestamp.string_to_timestamp(comparison_value)
                res = a == b
        return res

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        res = TimestampEquals._compare(variable, value)
        env.stack.append(res)


class TimestampEqualsPath(TimestampEquals):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.TimestampEqualsPath)

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        inp = env.stack[-1]
        comp_value = extract_json(value, inp)
        res = TimestampEqualsPath._compare(variable, comp_value)
        env.stack.append(res)


class TimestampGreaterThan(Operator):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.TimestampGreaterThan)

    @staticmethod
    def _compare(variable: Any, comparison_value: Any) -> bool:
        res = False
        if isinstance(variable, str):
            a = IsTimestamp.string_to_timestamp(variable)
            if a is not None:
                b = IsTimestamp.string_to_timestamp(comparison_value)
                res = a > b
        return res

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        res = TimestampGreaterThan._compare(variable, value)
        env.stack.append(res)


class TimestampGreaterThanPath(TimestampGreaterThan):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.TimestampGreaterThanPath)

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        inp = env.stack[-1]
        comp_value = extract_json(value, inp)
        res = TimestampGreaterThanPath._compare(variable, comp_value)
        env.stack.append(res)


class TimestampGreaterThanEquals(Operator):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.TimestampGreaterThanEquals)

    @staticmethod
    def _compare(variable: Any, comparison_value: Any) -> bool:
        res = False
        if isinstance(variable, str):
            a = IsTimestamp.string_to_timestamp(variable)
            if a is not None:
                b = IsTimestamp.string_to_timestamp(comparison_value)
                res = a >= b
        return res

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        res = TimestampGreaterThanEquals._compare(variable, value)
        env.stack.append(res)


class TimestampGreaterThanEqualsPath(TimestampGreaterThanEquals):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.TimestampGreaterThanEqualsPath)

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        inp = env.stack[-1]
        comp_value = extract_json(value, inp)
        res = TimestampGreaterThanEqualsPath._compare(variable, comp_value)
        env.stack.append(res)


class TimestampLessThan(Operator):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.TimestampLessThan)

    @staticmethod
    def _compare(variable: Any, comparison_value: Any) -> bool:
        res = False
        if isinstance(variable, str):
            a = IsTimestamp.string_to_timestamp(variable)
            if a is not None:
                b = IsTimestamp.string_to_timestamp(comparison_value)
                res = a < b
        return res

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        res = TimestampLessThan._compare(variable, value)
        env.stack.append(res)


class TimestampLessThanPath(TimestampLessThan):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.TimestampLessThanPath)

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        inp = env.stack[-1]
        comp_value = extract_json(value, inp)
        res = TimestampLessThanPath._compare(variable, comp_value)
        env.stack.append(res)


class TimestampLessThanEquals(Operator):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.TimestampLessThanEquals)

    @staticmethod
    def _compare(variable: Any, comparison_value: Any) -> bool:
        res = False
        if isinstance(variable, str):
            a = IsTimestamp.string_to_timestamp(variable)
            if a is not None:
                b = IsTimestamp.string_to_timestamp(comparison_value)
                res = a <= b
        return res

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        res = TimestampLessThanEquals._compare(variable, value)
        env.stack.append(res)


class TimestampLessThanEqualsPath(TimestampLessThanEquals):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.TimestampLessThanEqualsPath)

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        inp = env.stack[-1]
        comp_value = extract_json(value, inp)
        res = TimestampLessThanEqualsPath._compare(variable, comp_value)
        env.stack.append(res)



import json
import os
from datetime import datetime
from unittest import SkipTest, mock

import boto3

from moto import mock_aws, settings


@mock_aws
def test_create_partner_event_bus():
    client_account = "111122223333"
    client = boto3.client("events", "us-east-1")
    client.create_partner_event_source(
        Name="mypartner/actions/action1", Account=client_account
    )
    resp = client.describe_partner_event_source(Name="mypartner/actions/action1")
    assert (
        resp["Arn"]
        == "arn:aws:events:us-east-1::event-source/aws.partner/mypartner/actions/action1"
    )
    assert resp["Name"] == "mypartner/actions/action1"


@mock_aws
def test_describe_partner_event_busses():
    if settings.TEST_SERVER_MODE:
        raise SkipTest("Can't change accounts easily in ServerMode")
    # Having a separate partner account isn't 'strictly necessary - we could do that from the main account
    # But it makes it more obvious for the reader that we're accessing different accounts IMO
    partner_account = "111122223333"
    client_account = "444455556666"
    client = boto3.client("events", "us-east-1")
    name = "mypartner/actions/action1"
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": partner_account}):
        client.create_partner_event_source(Name=name, Account=client_account)

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": client_account}):
        resp = client.describe_event_source(Name=name)
        assert resp["Name"] == name
        assert resp["CreatedBy"] == "mypartner"
        assert resp["State"] == "ACTIVE"

        client.create_event_bus(Name=name, EventSourceName=name)
        resp = client.describe_event_bus(Name=name)
        assert resp["Name"] == name

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": partner_account}):
        client.delete_partner_event_source(Name=name, Account=client_account)

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": client_account}):
        resp = client.describe_event_source(Name=name)
        assert resp["State"] == "DELETED"


@mock_aws
def test_put_partner_events():
    if settings.TEST_SERVER_MODE:
        raise SkipTest("Can't change accounts easily in ServerMode")

    partner_account = "111122223333"
    client_account = "444455556666"
    events = boto3.client("events", "us-east-1")
    logs = boto3.client("logs", region_name="us-east-1")
    name = "mypartner/actions/action1"
    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": partner_account}):
        events.create_partner_event_source(Name=name, Account=client_account)

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": client_account}):
        events.create_event_bus(Name=name, EventSourceName=name)

        log_group_name = "/test-group"
        rule_name = "test-rule"
        logs.create_log_group(logGroupName=log_group_name)
        log_group_arn = (
            f"arn:aws:logs:us-east-1:{client_account}:log-group:{log_group_name}"
        )
        events.put_rule(
            Name=rule_name,
            EventPattern=json.dumps({"account": [client_account]}),
            State="ENABLED",
        )
        events.put_targets(
            Rule=rule_name,
            Targets=[{"Id": "logs", "Arn": log_group_arn}],
        )

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": partner_account}):
        resp = events.put_partner_events(
            Entries=[
                {
                    "Time": datetime.now(),
                    "Source": name,
                    "DetailType": "test-detail-type",
                    "Detail": json.dumps({"foo": "123", "bar": "123"}),
                }
            ]
        )
        assert resp["FailedEntryCount"] == 0

    with mock.patch.dict(os.environ, {"MOTO_ACCOUNT_ID": client_account}):
        log_events = logs.filter_log_events(logGroupName=log_group_name)["events"]
        assert len(log_events) == 1
        assert "test-detail-type" in log_events[0]["message"]



from typing import Optional

from moto.stepfunctions.parser.api import (
    HistoryEventType,
)
from moto.stepfunctions.parser.asl.component.common.parargs import Parameters, Parargs
from moto.stepfunctions.parser.asl.component.common.path.result_path import ResultPath
from moto.stepfunctions.parser.asl.component.state.state import CommonStateField
from moto.stepfunctions.parser.asl.component.state.state_pass.result import Result
from moto.stepfunctions.parser.asl.component.state.state_props import StateProps
from moto.stepfunctions.parser.asl.eval.environment import Environment


class StatePass(CommonStateField):
    def __init__(self):
        super(StatePass, self).__init__(
            state_entered_event_type=HistoryEventType.PassStateEntered,
            state_exited_event_type=HistoryEventType.PassStateExited,
        )

        # Result (Optional)
        # Refers to the output of a virtual state_task that is passed on to the next state. If you include the ResultPath
        # field in your state machine definition, Result is placed as specified by ResultPath and passed on to the
        self.result: Optional[Result] = None

        # ResultPath (Optional)
        # Specifies where to place the output (relative to the input) of the virtual state_task specified in Result. The input
        # is further filtered as specified by the OutputPath field (if present) before being used as the state's output.
        self.result_path: Optional[ResultPath] = None

        # Parameters (Optional)
        # Creates a collection of key-value pairs that will be passed as input. You can specify Parameters as a static
        # value or select from the input using a path.
        self.parameters: Optional[Parameters] = None

    def from_state_props(self, state_props: StateProps) -> None:
        super(StatePass, self).from_state_props(state_props)
        self.result = state_props.get(Result)
        self.result_path = state_props.get(ResultPath) or ResultPath(
            result_path_src=ResultPath.DEFAULT_PATH
        )
        self.parameters = state_props.get(Parargs)

    def _eval_state(self, env: Environment) -> None:
        if self.parameters:
            self.parameters.eval(env=env)

        if self.result:
            self.result.eval(env=env)

        if not self._is_language_query_jsonpath():
            output_value = env.stack[-1]
            env.states.set_result(output_value)

        if self.assign_decl:
            self.assign_decl.eval(env=env)

        if self.result_path:
            self.result_path.eval(env)



import datetime
from unittest import SkipTest
from unittest.mock import ANY

import boto3
from dateutil.tz import tzutc
from freezegun import freeze_time

from moto import mock_aws, settings

GIVEN_DEVICE_NAME = "not-a-device-name"
PROVISION_DEVICE_PARAMS = {
    "Description": "not a device description",
    "Name": GIVEN_DEVICE_NAME,
    "NetworkingConfiguration": {
        "Ethernet0": {
            "ConnectionType": "STATIC_IP",
            "StaticIpConnectionInfo": {
                "DefaultGateway": "192.168.1.1",
                "Dns": [
                    "8.8.8.8",
                ],
                "IpAddress": "192.168.1.10",
                "Mask": "255.255.255.0",
            },
        },
        "Ethernet1": {
            "ConnectionType": "dhcp",
        },
        "Ntp": {
            "NtpServers": [
                "0.pool.ntp.org",
                "1.pool.ntp.org",
                "0.fr.pool.ntp.org",
            ]
        },
    },
    "Tags": {"Key": "test-key", "Value": "test-value"},
}


@mock_aws
def test_create_application_instance() -> None:
    # Given
    panorama_client = boto3.client("panorama", "eu-west-1")
    response_device_creation = panorama_client.provision_device(
        **PROVISION_DEVICE_PARAMS
    )

    # When
    response = panorama_client.create_application_instance(
        DefaultRuntimeContextDevice=response_device_creation["DeviceId"],
        Description="not a description",
        ManifestOverridesPayload={
            "PayloadData": "dumped_payload_data",
        },
        ManifestPayload={
            "PayloadData": "dumped_manifest_payload_data",
        },
        Name="not-a-name",
        RuntimeRoleArn="not-an-arn",
        Tags={
            "Key": "value",
        },
    )
    # Then
    assert isinstance(response["ApplicationInstanceId"], str)
    assert response["ApplicationInstanceId"].startswith("applicationInstance-")
    assert len(response["ApplicationInstanceId"]) == 42


@mock_aws
def test_describe_application_instance() -> None:
    # Given
    panorama_client = boto3.client("panorama", "eu-west-1")
    response_device_creation = panorama_client.provision_device(
        **PROVISION_DEVICE_PARAMS
    )
    given_application_instance_name = "not-a-name"
    given_application_instance_arn = "not-an-arn"
    response_created = panorama_client.create_application_instance(
        DefaultRuntimeContextDevice=response_device_creation["DeviceId"],
        Description="not a description",
        ManifestOverridesPayload={
            "PayloadData": "dumped_payload_data",
        },
        ManifestPayload={
            "PayloadData": "dumped_manifest_payload_data",
        },
        Name=given_application_instance_name,
        RuntimeRoleArn=given_application_instance_arn,
        Tags={
            "Key": "value",
        },
    )

    # When
    response = panorama_client.describe_application_instance(
        ApplicationInstanceId=response_created["ApplicationInstanceId"]
    )

    # Then
    assert (
        response["ApplicationInstanceId"] == response_created["ApplicationInstanceId"]
    )
    assert response.get("ApplicationInstanceIdToReplace") is None
    assert response["Arn"].startswith(
        "arn:aws:panorama:eu-west-1:123456789012:application-instance/"
    )
    assert response["Arn"].endswith(response_created["ApplicationInstanceId"])
    assert isinstance(response["CreatedTime"], datetime.datetime)
    assert (
        response["DefaultRuntimeContextDevice"] == response_device_creation["DeviceId"]
    )
    assert response["DefaultRuntimeContextDeviceName"] == "not-a-device-name"
    assert response["Description"] == "not a description"
    assert response["HealthStatus"] == "RUNNING"
    assert isinstance(response["LastUpdatedTime"], datetime.datetime)
    assert response["Name"] == given_application_instance_name
    assert response["RuntimeRoleArn"] == given_application_instance_arn
    assert response["Status"] == "DEPLOYMENT_SUCCEEDED"


@mock_aws
def test_create_application_instance_should_set_created_time() -> None:
    # Given
    if settings.TEST_SERVER_MODE:
        raise SkipTest("Can't use ManagedState in ServerMode")
    panorama_client = boto3.client("panorama", "eu-west-1")
    response_device_creation = panorama_client.provision_device(
        **PROVISION_DEVICE_PARAMS
    )
    given_application_instance_name = "not-a-name"
    given_application_instance_arn = "not-an-arn"
    with freeze_time("2020-01-01 12:00:00"):
        response_created = panorama_client.create_application_instance(
            DefaultRuntimeContextDevice=response_device_creation["DeviceId"],
            Description="not a description",
            ManifestOverridesPayload={
                "PayloadData": "dumped_payload_data",
            },
            ManifestPayload={
                "PayloadData": "dumped_manifest_payload_data",
            },
            Name=given_application_instance_name,
            RuntimeRoleArn=given_application_instance_arn,
            Tags={
                "Key": "value",
            },
        )

    # When
    response = panorama_client.describe_application_instance(
        ApplicationInstanceId=response_created["ApplicationInstanceId"]
    )

    # Then
    assert response["CreatedTime"] == datetime.datetime(
        2020, 1, 1, 12, 0, 0, tzinfo=tzutc()
    )
    assert response["LastUpdatedTime"] == datetime.datetime(
        2020, 1, 1, 12, 0, 0, tzinfo=tzutc()
    )


@mock_aws
def test_describe_application_instance_details() -> None:
    # Given
    panorama_client = boto3.client("panorama", "eu-west-1")
    given_application_instance_name = "not-a-name"
    given_application_instance_arn = "not-an-arn"
    response_device_creation = panorama_client.provision_device(
        **PROVISION_DEVICE_PARAMS
    )
    given_manifest_payload = {
        "PayloadData": "dumped_manifest_payload_data",
    }
    given_manifest_overrides_payload = {
        "PayloadData": "dumped_payload_data",
    }
    response_created = panorama_client.create_application_instance(
        DefaultRuntimeContextDevice=response_device_creation["DeviceId"],
        Description="not a description",
        ManifestOverridesPayload=given_manifest_overrides_payload,
        ManifestPayload=given_manifest_payload,
        Name=given_application_instance_name,
        RuntimeRoleArn=given_application_instance_arn,
        Tags={
            "Key": "value",
        },
    )

    # When
    response = panorama_client.describe_application_instance_details(
        ApplicationInstanceId=response_created["ApplicationInstanceId"]
    )

    # Then
    assert (
        response["ApplicationInstanceId"] == response_created["ApplicationInstanceId"]
    )
    assert response.get("ApplicationInstanceIdToReplace") is None
    assert isinstance(response["CreatedTime"], datetime.datetime)
    assert (
        response["DefaultRuntimeContextDevice"] == response_device_creation["DeviceId"]
    )
    assert response["Description"] == "not a description"
    assert response["Name"] == given_application_instance_name
    assert response["ManifestPayload"] == given_manifest_payload
    assert response["ManifestOverridesPayload"] == given_manifest_overrides_payload


@mock_aws
def test_list_application_instances() -> None:
    # Given
    panorama_client = boto3.client("panorama", "eu-west-1")
    given_application_instance_name = "not-a-name"
    given_application_instance_arn = "not-an-arn"
    given_device_name = "not-a-device-name"
    response_device_creation = panorama_client.provision_device(
        **PROVISION_DEVICE_PARAMS
    )
    response_created_1 = panorama_client.create_application_instance(
        DefaultRuntimeContextDevice=response_device_creation["DeviceId"],
        Description="not a description",
        ManifestOverridesPayload={
            "PayloadData": "dumped_payload_data",
        },
        ManifestPayload={
            "PayloadData": "dumped_manifest_payload_data",
        },
        Name=given_application_instance_name,
        RuntimeRoleArn=given_application_instance_arn,
        Tags={
            "Key": "value",
        },
    )
    response_created_2 = panorama_client.create_application_instance(
        ApplicationInstanceIdToReplace=response_created_1["ApplicationInstanceId"],
        DefaultRuntimeContextDevice=response_device_creation["DeviceId"],
        Description="not a description",
        ManifestOverridesPayload={
            "PayloadData": "dumped_payload_data",
        },
        ManifestPayload={
            "PayloadData": "dumped_manifest_payload_data",
        },
        Name=given_application_instance_name,
        RuntimeRoleArn=given_application_instance_arn,
        Tags={
            "Key": "value",
        },
    )

    # When
    response_1 = panorama_client.list_application_instances(
        DeviceId=response_device_creation["DeviceId"],
        MaxResults=1,
    )
    response_2 = panorama_client.list_application_instances(
        DeviceId=response_device_creation["DeviceId"],
        MaxResults=1,
        NextToken=response_1["NextToken"],
    )

    # Then
    assert len(response_1["ApplicationInstances"]) == 1
    assert (
        response_1["ApplicationInstances"][0]["ApplicationInstanceId"]
        == response_created_1["ApplicationInstanceId"]
    )
    assert response_1["ApplicationInstances"][0]["Arn"].startswith(
        "arn:aws:panorama:eu-west-1:123456789012:application-instance/"
    )
    assert response_1["ApplicationInstances"][0]["Arn"].endswith(
        response_created_1["ApplicationInstanceId"]
    )
    assert isinstance(
        response_1["ApplicationInstances"][0]["CreatedTime"], datetime.datetime
    )
    assert (
        response_1["ApplicationInstances"][0]["DefaultRuntimeContextDevice"]
        == response_device_creation["DeviceId"]
    )
    assert (
        response_1["ApplicationInstances"][0]["DefaultRuntimeContextDeviceName"]
        == given_device_name
    )
    assert response_1["ApplicationInstances"][0]["Description"] == "not a description"
    assert response_1["ApplicationInstances"][0]["HealthStatus"] == "RUNNING"
    assert response_1["ApplicationInstances"][0]["Name"], response_created_1["Name"]
    assert response_1["ApplicationInstances"][0]["RuntimeContextStates"] == [
        {
            "DesiredState": "RUNNING",
            "DeviceReportedStatus": "RUNNING",
            "DeviceReportedTime": ANY,
            "RuntimeContextName": "string",
        },
        {
            "DesiredState": "REMOVED",
            "DeviceReportedStatus": "REMOVAL_IN_PROGRESS",
            "DeviceReportedTime": ANY,
            "RuntimeContextName": "string",
        },
    ]
    assert response_1["ApplicationInstances"][0]["Status"] == "REMOVAL_SUCCEEDED"
    assert response_1["ApplicationInstances"][0]["StatusDescription"] == "string"
    assert response_1["ApplicationInstances"][0]["Tags"] == {"Key": "value"}
    assert response_1["NextToken"] is not None

    assert len(response_2["ApplicationInstances"]) == 1
    assert (
        response_2["ApplicationInstances"][0]["ApplicationInstanceId"]
        == response_created_2["ApplicationInstanceId"]
    )
    assert "NextToken" not in response_2
    assert response_2["ApplicationInstances"][0]["RuntimeContextStates"] == [
        {
            "DesiredState": "RUNNING",
            "DeviceReportedStatus": "RUNNING",
            "DeviceReportedTime": ANY,
            "RuntimeContextName": "string",
        }
    ]


@mock_aws
def test_list_application_should_return_only_filtered_results_if_status_filter_used() -> (
    None
):
    # Given
    panorama_client = boto3.client("panorama", "eu-west-1")
    given_application_instance_name = "not-a-name"
    given_application_instance_name_2 = "not-a-name-2"
    given_application_instance_arn = "not-an-arn"
    response_device_creation = panorama_client.provision_device(
        **PROVISION_DEVICE_PARAMS
    )
    response_created_1 = panorama_client.create_application_instance(
        DefaultRuntimeContextDevice=response_device_creation["DeviceId"],
        Description="not a description",
        ManifestOverridesPayload={
            "PayloadData": "dumped_payload_data",
        },
        ManifestPayload={
            "PayloadData": "dumped_manifest_payload_data",
        },
        Name=given_application_instance_name,
        RuntimeRoleArn=given_application_instance_arn,
        Tags={
            "Key": "value",
        },
    )
    response_created_2 = panorama_client.create_application_instance(
        ApplicationInstanceIdToReplace=response_created_1["ApplicationInstanceId"],
        DefaultRuntimeContextDevice=response_device_creation["DeviceId"],
        Description="not a description",
        ManifestOverridesPayload={
            "PayloadData": "dumped_payload_data",
        },
        ManifestPayload={
            "PayloadData": "dumped_manifest_payload_data",
        },
        Name=given_application_instance_name_2,
        RuntimeRoleArn=given_application_instance_arn,
        Tags={
            "Key": "value",
        },
    )

    # When
    response_deployed = panorama_client.list_application_instances(
        DeviceId=response_device_creation["DeviceId"],
        MaxResults=123,
        StatusFilter="DEPLOYMENT_SUCCEEDED",
    )

    response_removed = panorama_client.list_application_instances(
        DeviceId=response_device_creation["DeviceId"],
        MaxResults=123,
        StatusFilter="REMOVAL_SUCCEEDED",
    )

    # Then
    assert len(response_removed["ApplicationInstances"]) == 1
    assert (
        response_removed["ApplicationInstances"][0]["ApplicationInstanceId"]
        == response_created_1["ApplicationInstanceId"]
    )
    assert len(response_deployed["ApplicationInstances"]) == 1
    assert (
        response_deployed["ApplicationInstances"][0]["ApplicationInstanceId"]
        == response_created_2["ApplicationInstanceId"]
    )



from typing import Any, Callable, Dict, Optional, Set

from moto.stepfunctions.parser.asl.component.state.exec.state_task.credentials import (
    ComputedCredentials,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceCondition,
    ResourceRuntimePart,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_callback import (
    StateTaskServiceCallback,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.utils.boto_client import boto_client_for

_SUPPORTED_INTEGRATION_PATTERNS: Set[ResourceCondition] = {
    ResourceCondition.WaitForTaskToken,
    ResourceCondition.Sync,
}

_SUPPORTED_API_PARAM_BINDINGS: Dict[str, Set[str]] = {
    "runtask": {
        "Cluster",
        "Group",
        "LaunchType",
        "NetworkConfiguration",
        "Overrides",
        "PlacementConstraints",
        "PlacementStrategy",
        "PlatformVersion",
        "PropagateTags",
        "TaskDefinition",
        "EnableExecuteCommand",
    }
}

_STARTED_BY_PARAMETER_RAW_KEY: str = "StartedBy"
_STARTED_BY_PARAMETER_VALUE: str = "AWS Step Functions"


class StateTaskServiceEcs(StateTaskServiceCallback):
    def __init__(self):
        super().__init__(supported_integration_patterns=_SUPPORTED_INTEGRATION_PATTERNS)

    def _get_supported_parameters(self) -> Optional[Set[str]]:
        return _SUPPORTED_API_PARAM_BINDINGS.get(self.resource.api_action.lower())

    def _before_eval_execution(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        raw_parameters: dict,
        task_credentials: ComputedCredentials,
    ) -> None:
        if self.resource.condition == ResourceCondition.Sync:
            raw_parameters[_STARTED_BY_PARAMETER_RAW_KEY] = _STARTED_BY_PARAMETER_VALUE
        super()._before_eval_execution(
            env=env,
            resource_runtime_part=resource_runtime_part,
            raw_parameters=raw_parameters,
            task_credentials=task_credentials,
        )

    def _eval_service_task(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ):
        service_name = self._get_boto_service_name()
        api_action = self._get_boto_service_action()
        ecs_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service=service_name,
            credentials=task_credentials,
        )
        response = getattr(ecs_client, api_action)(**normalised_parameters)
        response.pop("ResponseMetadata", None)

        # AWS outputs the description of the task, not the output of run_task.
        if self._get_boto_service_action() == "run_task":
            self._normalise_response(response=response, service_action_name="run_task")
            cluster_arn: str = response["Tasks"][0]["ClusterArn"]
            task_arn: str = response["Tasks"][0]["TaskArn"]
            describe_tasks_output = ecs_client.describe_tasks(
                cluster=cluster_arn, tasks=[task_arn]
            )
            describe_tasks_output.pop("ResponseMetadata", None)
            self._normalise_response(
                response=describe_tasks_output, service_action_name="describe_tasks"
            )
            env.stack.append(describe_tasks_output)
            return

        env.stack.append(response)

    def _build_sync_resolver(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ) -> Callable[[], Optional[Any]]:
        ecs_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service="ecs",
            credentials=task_credentials,
        )
        submission_output: dict = env.stack.pop()
        task_arn: str = submission_output["Tasks"][0]["TaskArn"]
        cluster_arn: str = submission_output["Tasks"][0]["ClusterArn"]

        def _sync_resolver() -> Optional[dict]:
            describe_tasks_output = ecs_client.describe_tasks(
                cluster=cluster_arn, tasks=[task_arn]
            )
            last_status: str = describe_tasks_output["tasks"][0]["lastStatus"]

            if last_status == "STOPPED":
                self._normalise_response(
                    response=describe_tasks_output, service_action_name="describe_tasks"
                )
                return describe_tasks_output["Tasks"][0]  # noqa

            return None

        return _sync_resolver



"""Test different server responses."""

import moto.server as server


def test_elasticbeanstalk_describe():
    backend = server.create_backend_app("elasticbeanstalk")
    test_client = backend.test_client()

    data = "Action=DescribeApplications"
    headers = {"Host": "elasticbeanstalk.us-east-1.amazonaws.com"}
    resp = test_client.post("/", data=data, headers=headers)

    assert resp.status_code == 200
    assert "<Applications></Applications>" in str(resp.data)



import abc
from typing import Final

from moto.stepfunctions.parser.api import ExecutionFailedEventDetails, HistoryEventType
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value_terminal import (
    JSONataTemplateValueTerminalExpression,
)
from moto.stepfunctions.parser.asl.component.common.variable_sample import (
    VariableSample,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.json_path import extract_json


class MaxItemsDecl(EvalComponent, abc.ABC):
    """
    "MaxItems": Limits the number of data items passed to the Map state. For example, suppose that you provide a
    CSV file that contains 1000 rows and specify a limit of 100. Then, the interpreter passes only 100 rows to the
    Map state. The Map state processes items in sequential order, starting after the header row.
    Currently, you can specify a limit of up to 100,000,000
    """

    MAX_VALUE: Final[int] = 100_000_000

    def _clip_value(self, value: int) -> int:
        if value == 0:
            return self.MAX_VALUE
        return min(value, self.MAX_VALUE)

    @abc.abstractmethod
    def _get_value(self, env: Environment) -> int: ...

    def _eval_body(self, env: Environment) -> None:
        max_items: int = self._get_value(env=env)
        max_items = self._clip_value(max_items)
        env.stack.append(max_items)


class MaxItems(MaxItemsDecl):
    max_items: Final[int]

    def __init__(self, max_items: int = MaxItemsDecl.MAX_VALUE):
        if max_items < 0 or max_items > MaxItems.MAX_VALUE:
            raise ValueError(
                f"MaxItems value MUST be a non-negative integer "
                f"non greater than '{MaxItems.MAX_VALUE}', got '{max_items}'."
            )
        self.max_items = max_items

    def _get_value(self, env: Environment) -> int:
        return self.max_items


class MaxItemsJSONata(MaxItemsDecl):
    jsonata_template_value_terminal_expression: Final[
        JSONataTemplateValueTerminalExpression
    ]

    def __init__(
        self,
        jsonata_template_value_terminal_expression: JSONataTemplateValueTerminalExpression,
    ):
        super().__init__()
        self.jsonata_template_value_terminal_expression = (
            jsonata_template_value_terminal_expression
        )

    def _get_value(self, env: Environment) -> int:
        # TODO: add snapshot tests to verify AWS's behaviour about non integer values.
        self.jsonata_template_value_terminal_expression.eval(env=env)
        max_items: int = int(env.stack.pop())
        return max_items


class MaxItemsPathVar(MaxItemsDecl):
    variable_sample: VariableSample

    def __init__(self, variable_sample: VariableSample):
        super().__init__()
        self.variable_sample = variable_sample

    def _get_value(self, env: Environment) -> int:
        self.variable_sample.eval(env=env)
        # TODO: add snapshot tests to verify AWS's behaviour about non integer values.
        max_items: int = int(env.stack.pop())
        return max_items


class MaxItemsPath(MaxItemsDecl):
    """
    "MaxItemsPath": computes a MaxItems value equal to the reference path it points to.
    """

    def __init__(self, path: str):
        self.path: Final[str] = path

    def _validate_value(self, env: Environment, value: int) -> None:
        if not isinstance(value, int):
            # TODO: Note, this error appears to be validated at a earlier stage in AWS Step Functions, unlike the
            #  negative integer check that is validated at this exact depth.
            error_typ = StatesErrorNameType.StatesItemReaderFailed
            raise FailureEventException(
                failure_event=FailureEvent(
                    env=env,
                    error_name=StatesErrorName(typ=error_typ),
                    event_type=HistoryEventType.ExecutionFailed,
                    event_details=EventDetails(
                        executionFailedEventDetails=ExecutionFailedEventDetails(
                            error=error_typ.to_name(),
                            cause=(
                                f"The MaxItemsPath field refers to value '{value}' "
                                f"which is not a valid integer: {self.path}"
                            ),
                        )
                    ),
                )
            )
        if value < 0:
            error_typ = StatesErrorNameType.StatesItemReaderFailed
            raise FailureEventException(
                failure_event=FailureEvent(
                    env=env,
                    error_name=StatesErrorName(typ=error_typ),
                    event_type=HistoryEventType.MapRunFailed,
                    event_details=EventDetails(
                        executionFailedEventDetails=ExecutionFailedEventDetails(
                            error=error_typ.to_name(),
                            cause="field MaxItems must be positive",
                        )
                    ),
                )
            )

    def _get_value(self, env: Environment) -> int:
        inp = env.stack[-1]
        max_items = extract_json(self.path, inp)
        self._validate_value(env=env, value=max_items)
        return max_items



"""Exceptions raised by the cloudhsmv2 service."""

from moto.core.exceptions import JsonRESTError


class CloudHSMv2ClientError(JsonRESTError):
    """Base class for CloudHSMv2 errors."""

    code = 400


class ResourceNotFoundException(CloudHSMv2ClientError):
    def __init__(self, message: str):
        super().__init__("ResourceNotFoundException", message)


class InvalidRequestException(CloudHSMv2ClientError):
    def __init__(self, message: str):
        super().__init__("InvalidRequestException", message)


class ClientError(CloudHSMv2ClientError):
    def __init__(self, error_type: str, message: str):
        super().__init__(error_type, message)



from typing import Set

from moto.stepfunctions.parser.asl.antlr.runtime.ASLParser import ASLParser
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ActivityResource,
    Resource,
    ServiceResource,
)
from moto.stepfunctions.parser.asl.component.state.state_type import StateType
from moto.stepfunctions.parser.asl.parse.test_state.asl_parser import (
    TestStateAmazonStateLanguageParser,
)
from moto.stepfunctions.parser.asl.static_analyser.static_analyser import StaticAnalyser


class TestStateStaticAnalyser(StaticAnalyser):
    _SUPPORTED_STATE_TYPES: Set[StateType] = {
        StateType.Task,
        StateType.Pass,
        StateType.Wait,
        StateType.Choice,
        StateType.Succeed,
        StateType.Fail,
    }

    def analyse(self, definition) -> None:
        _, parser_rule_context = TestStateAmazonStateLanguageParser.parse(definition)
        self.visit(parser_rule_context)

    def visitState_type(self, ctx: ASLParser.State_typeContext) -> None:
        state_type_value: int = ctx.children[0].symbol.type
        state_type = StateType(state_type_value)
        if state_type not in self._SUPPORTED_STATE_TYPES:
            raise ValueError(
                f"Unsupported state type for TestState runs '{state_type}'."
            )

    def visitResource_decl(self, ctx: ASLParser.Resource_declContext) -> None:
        resource_str: str = ctx.keyword_or_string().getText()[1:-1]
        resource = Resource.from_resource_arn(resource_str)

        if isinstance(resource, ActivityResource):
            raise ValueError(
                f"ActivityResources are not supported for TestState runs {resource_str}."
            )

        if isinstance(resource, ServiceResource):
            if resource.condition is not None:
                raise ValueError(
                    f"Service integration patterns are not supported for TestState runs {resource_str}."
                )



import base64
import hashlib
import hmac
import io
import os
import struct
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from enum import Enum
from typing import Any, Dict, List, Tuple

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives._asymmetric import AsymmetricPadding
from cryptography.hazmat.primitives.asymmetric import ec, padding, rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from moto.moto_api._internal import mock_random

from .exceptions import (
    AccessDeniedException,
    InvalidCiphertextException,
    NotFoundException,
    ValidationException,
)

MASTER_KEY_LEN = 32
KEY_ID_LEN = 36
IV_LEN = 12
TAG_LEN = 16
HEADER_LEN = KEY_ID_LEN + IV_LEN + TAG_LEN
# NOTE: This is just a simple binary format. It is not what KMS actually does.
CIPHERTEXT_HEADER_FORMAT = f">{KEY_ID_LEN}s{IV_LEN}s{TAG_LEN}s"
Ciphertext = namedtuple("Ciphertext", ("key_id", "iv", "ciphertext", "tag"))

RESERVED_ALIASE_TARGET_KEY_IDS = {
    # NOTE: These would technically differ across account, but in that they are
    # out of customer control, testing that they are different would be redundant.
    "alias/aws/acm": "4f58743d-e279-4214-9270-8cc28277958d",
    "alias/aws/dynamodb": "7e6aa0ea-15a4-4e72-8b32-58e46e776888",
    "alias/aws/ebs": "7adeb491-68c9-4a5b-86ec-a86ce5364094",
    "alias/aws/elasticfilesystem": "0ef0f111-cdc8-4dda-b0bc-bf625bd5f154",
    "alias/aws/es": "3c7c1880-c353-4cea-9866-d8bc12f05573",
    "alias/aws/glue": "90fd783f-e582-4cc2-a207-672ee67f8d58",
    "alias/aws/kinesisvideo": "7fd4bff3-6eb7-4283-8f11-a7e0a793a181",
    "alias/aws/lambda": "ff9c4f27-2f29-4d9b-bf38-02f88b52a70c",
    "alias/aws/rds": "f5f30938-abed-41a2-a0f6-5482d02a2489",
    "alias/aws/redshift": "dcdae9aa-593a-4e0b-9153-37325591901f",
    "alias/aws/s3": "8c3faf07-f43c-4d11-abdb-9183079214c7",
    "alias/aws/secretsmanager": "fee5173a-3972-428e-ae73-cd4c2a371222",
    "alias/aws/ssm": "cb3f6250-5078-48c0-a75f-0290bf47694e",
    "alias/aws/xray": "e9b758eb-6230-4744-93d1-ad3b7d71f2f6",
}

RESERVED_ALIASES = list(RESERVED_ALIASE_TARGET_KEY_IDS.keys())


class KeySpec(str, Enum):
    # Asymmetric key specs
    RSA_2048 = "RSA_2048"
    RSA_3072 = "RSA_3072"
    RSA_4096 = "RSA_4096"
    ECC_NIST_P256 = "ECC_NIST_P256"
    ECC_SECG_P256K1 = "ECC_SECG_P256K1"
    ECC_NIST_P384 = "ECC_NIST_P384"
    ECC_NIST_P521 = "ECC_NIST_P521"
    SM2 = "SM2"  # China Regions only
    # Symmetric key specs
    SYMMETRIC_DEFAULT = "SYMMETRIC_DEFAULT"
    HMAC_224 = "HMAC_224"
    HMAC_256 = "HMAC_256"
    HMAC_284 = "HMAC_384"
    HMAC_512 = "HMAC_512"

    @classmethod
    def key_specs(self) -> List[str]:
        return sorted([item.value for item in KeySpec])

    @classmethod
    def rsa_key_specs(self) -> List[str]:
        return [spec for spec in self.key_specs() if spec.startswith("RSA")]

    @classmethod
    def ecc_key_specs(self) -> List[str]:
        return [spec for spec in self.key_specs() if spec.startswith("ECC")]

    @classmethod
    def hmac_key_specs(self) -> List[str]:
        return [spec for spec in self.key_specs() if spec.startswith("HMAC")]


class SigningAlgorithm(str, Enum):
    # sigingin algorithms for RSA key spec
    RSASSA_PSS_SHA_256 = "RSASSA_PSS_SHA_256"
    RSASSA_PSS_SHA_384 = "RSASSA_PSS_SHA_384"
    RSASSA_PSS_SHA_512 = "RSASSA_PSS_SHA_512"
    RSASSA_PKCS1_V1_5_SHA_256 = "RSASSA_PKCS1_V1_5_SHA_256"
    RSASSA_PKCS1_V1_5_SHA_384 = "RSASSA_PKCS1_V1_5_SHA_384"
    RSASSA_PKCS1_V1_5_SHA_512 = "RSASSA_PKCS1_V1_5_SHA_512"
    # sigining algorithms for ECC_NIST_P256, P256K1 spec
    ECDSA_SHA_256 = "ECDSA_SHA_256"
    # siginging algorithm for ECC_NIST_P384
    ECDSA_SHA_384 = "ECDSA_SHA_384"
    # sigining algorithm for ECC_NIST_P512
    ECDSA_SHA_512 = "ECDSA_SHA_512"
    # sigining algorithm for SM2
    SM2DSA = "SM2DSA"

    @classmethod
    def signing_algorithms(self) -> List[str]:
        return sorted([item.value for item in SigningAlgorithm])

    @classmethod
    def rsa_signing_algorithms(self) -> List[str]:
        return [algo for algo in self.signing_algorithms() if algo.startswith("RSASSA")]

    @classmethod
    def ecc_signing_algorithms(self) -> List[str]:
        return [algo for algo in self.signing_algorithms() if algo.startswith("ECDSA")]


def generate_key_id(multi_region: bool = False) -> str:
    key = str(mock_random.uuid4())
    # https://docs.aws.amazon.com/kms/latest/developerguide/multi-region-keys-overview.html
    # "Notice that multi-Region keys have a distinctive key ID that begins with mrk-. You can use the mrk- prefix to
    # identify MRKs programmatically."
    if multi_region:
        key = "mrk-" + key

    return key


def generate_data_key(number_of_bytes: int) -> bytes:
    """Generate a data key."""
    return os.urandom(number_of_bytes)


def generate_master_key() -> bytes:
    """Generate a master key."""
    return generate_data_key(MASTER_KEY_LEN)


class AbstractPrivateKey(metaclass=ABCMeta):
    @abstractmethod
    def sign(self, message: bytes, signing_algorithm: str) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def verify(self, message: bytes, signature: bytes, signing_algorithm: str) -> bool:
        raise NotImplementedError

    @abstractmethod
    def public_key(self) -> bytes:
        raise NotImplementedError


def validate_signing_algorithm(
    target_algorithm: str, valid_algorithms: List[str]
) -> None:
    if target_algorithm not in valid_algorithms:
        raise ValidationException(
            (
                "1 validation error detected: Value at 'signing_algorithm' failed"
                "to satisfy constraint: Member must satisfy enum value set: {valid_signing_algorithms}"
            ).format(valid_signing_algorithms=valid_algorithms)
        )


def validate_key_spec(target_key_spec: str, valid_key_specs: List[str]) -> None:
    if target_key_spec not in valid_key_specs:
        raise ValidationException(
            (
                "1 validation error detected: Value at 'key_spec' failed "
                "to satisfy constraint: Member must satisfy enum value set: {valid_key_specs}"
            ).format(valid_key_specs=valid_key_specs)
        )


class RSAPrivateKey(AbstractPrivateKey):
    # See https://docs.aws.amazon.com/kms/latest/cryptographic-details/crypto-primitives.html
    __supported_key_sizes = [2048, 3072, 4096]

    def __init__(self, key_size: int):
        if key_size not in self.__supported_key_sizes:
            raise ValidationException(
                (
                    "1 validation error detected: Value at 'key_size' failed "
                    "to satisfy constraint: Member must satisfy enum value set: {supported_key_sizes}"
                ).format(supported_key_sizes=self.__supported_key_sizes)
            )
        self.key_size = key_size
        self.private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=self.key_size
        )

    def __padding_and_hash_algorithm(
        self, signing_algorithm: str
    ) -> Tuple[AsymmetricPadding, hashes.HashAlgorithm]:
        if signing_algorithm == SigningAlgorithm.RSASSA_PSS_SHA_256:
            pad = padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()), salt_length=padding.PSS.MAX_LENGTH
            )  # type: AsymmetricPadding
            algorithm = hashes.SHA256()  # type: Any
        elif signing_algorithm == SigningAlgorithm.RSASSA_PSS_SHA_384:
            pad = padding.PSS(
                mgf=padding.MGF1(hashes.SHA384()), salt_length=padding.PSS.MAX_LENGTH
            )
            algorithm = hashes.SHA384()
        elif signing_algorithm == SigningAlgorithm.RSASSA_PSS_SHA_512:
            pad = padding.PSS(
                mgf=padding.MGF1(hashes.SHA512()), salt_length=padding.PSS.MAX_LENGTH
            )
            algorithm = hashes.SHA512()
        elif signing_algorithm == SigningAlgorithm.RSASSA_PKCS1_V1_5_SHA_256:
            pad = padding.PKCS1v15()
            algorithm = hashes.SHA256()
        elif signing_algorithm == SigningAlgorithm.RSASSA_PKCS1_V1_5_SHA_384:
            pad = padding.PKCS1v15()
            algorithm = hashes.SHA384()
        else:
            pad = padding.PKCS1v15()
            algorithm = hashes.SHA512()
        return pad, algorithm

    def sign(self, message: bytes, signing_algorithm: str) -> bytes:
        validate_signing_algorithm(
            signing_algorithm, SigningAlgorithm.rsa_signing_algorithms()
        )
        pad, hash_algorithm = self.__padding_and_hash_algorithm(signing_algorithm)
        return self.private_key.sign(message, pad, hash_algorithm)

    def verify(self, message: bytes, signature: bytes, signing_algorithm: str) -> bool:
        validate_signing_algorithm(
            signing_algorithm, SigningAlgorithm.rsa_signing_algorithms()
        )
        pad, hash_algorithm = self.__padding_and_hash_algorithm(signing_algorithm)
        public_key = self.private_key.public_key()
        try:
            public_key.verify(signature, message, pad, hash_algorithm)
            return True
        except InvalidSignature:
            return False

    def public_key(self) -> bytes:
        return self.private_key.public_key().public_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )


class ECDSAPrivateKey(AbstractPrivateKey):
    def __init__(self, key_spec: str):
        validate_key_spec(key_spec, KeySpec.ecc_key_specs())

        if key_spec == KeySpec.ECC_NIST_P256:
            curve = ec.SECP256R1()  # type: ec.EllipticCurve
            valid_signing_algorithms = ["ECDSA_SHA_256"]  # type: List[str]
        elif key_spec == KeySpec.ECC_SECG_P256K1:
            curve = ec.SECP256K1()
            valid_signing_algorithms = ["ECDSA_SHA_256"]
        elif key_spec == KeySpec.ECC_NIST_P384:
            curve = ec.SECP384R1()
            valid_signing_algorithms = ["ECDSA_SHA_384"]
        else:
            curve = ec.SECP521R1()
            valid_signing_algorithms = ["ECDSA_SHA_512"]

        self.private_key = ec.generate_private_key(curve)
        self.valid_signing_algorithms = valid_signing_algorithms

    def __hash_algorithm(self, signing_algorithm: str) -> hashes.HashAlgorithm:
        if signing_algorithm == SigningAlgorithm.ECDSA_SHA_256:
            algorithm = hashes.SHA256()  # type: Any
        elif signing_algorithm == SigningAlgorithm.ECDSA_SHA_384:
            algorithm = hashes.SHA384()
        else:
            algorithm = hashes.SHA512()
        return algorithm

    def sign(self, message: bytes, signing_algorithm: str) -> bytes:
        validate_signing_algorithm(signing_algorithm, self.valid_signing_algorithms)
        hash_algorithm = self.__hash_algorithm(signing_algorithm)
        return self.private_key.sign(message, ec.ECDSA(hash_algorithm))

    def verify(self, message: bytes, signature: bytes, signing_algorithm: str) -> bool:
        validate_signing_algorithm(signing_algorithm, self.valid_signing_algorithms)
        hash_algorithm = self.__hash_algorithm(signing_algorithm)
        public_key = self.private_key.public_key()
        try:
            public_key.verify(signature, message, ec.ECDSA(hash_algorithm))
            return True
        except InvalidSignature:
            return False

    def public_key(self) -> bytes:
        return self.private_key.public_key().public_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )


def generate_private_key(key_spec: str) -> AbstractPrivateKey:
    """Generate a private key to be used on asymmetric sign/verify."""
    if key_spec == KeySpec.RSA_2048:
        return RSAPrivateKey(key_size=2048)
    elif key_spec == KeySpec.RSA_3072:
        return RSAPrivateKey(key_size=3072)
    elif key_spec == KeySpec.RSA_4096:
        return RSAPrivateKey(key_size=4096)
    elif key_spec in KeySpec.ecc_key_specs():
        return ECDSAPrivateKey(key_spec)
    else:
        return RSAPrivateKey(key_size=2048)


def _serialize_ciphertext_blob(ciphertext: Ciphertext) -> bytes:
    """Serialize Ciphertext object into a ciphertext blob.

    NOTE: This is just a simple binary format. It is not what KMS actually does.
    """
    header = struct.pack(
        CIPHERTEXT_HEADER_FORMAT,
        ciphertext.key_id.encode("utf-8"),
        ciphertext.iv,
        ciphertext.tag,
    )
    return header + ciphertext.ciphertext


def _deserialize_ciphertext_blob(ciphertext_blob: bytes) -> Ciphertext:
    """Deserialize ciphertext blob into a Ciphertext object.

    NOTE: This is just a simple binary format. It is not what KMS actually does.
    """
    header = ciphertext_blob[:HEADER_LEN]
    ciphertext = ciphertext_blob[HEADER_LEN:]
    key_id, iv, tag = struct.unpack(CIPHERTEXT_HEADER_FORMAT, header)
    return Ciphertext(
        key_id=key_id.decode("utf-8"), iv=iv, ciphertext=ciphertext, tag=tag
    )


def _serialize_encryption_context(encryption_context: Dict[str, str]) -> bytes:
    """Serialize encryption context for use a AAD.

    NOTE: This is not necessarily what KMS does, but it retains the same properties.
    """
    aad = io.BytesIO()
    for key, value in sorted(encryption_context.items(), key=lambda x: x[0]):
        aad.write(key.encode("utf-8"))
        aad.write(value.encode("utf-8"))
    return aad.getvalue()


def encrypt(
    master_keys: Dict[str, Any],
    key_id: str,
    plaintext: bytes,
    encryption_context: Dict[str, str],
) -> bytes:
    """Encrypt data using a master key material.

    NOTE: This is not necessarily what KMS does, but it retains the same properties.

    NOTE: This function is NOT compatible with KMS APIs.
    :param dict master_keys: Mapping of a KmsBackend's known master keys
    :param str key_id: Key ID of moto master key
    :param bytes plaintext: Plaintext data to encrypt
    :param dict[str, str] encryption_context: KMS-style encryption context
    :returns: Moto-structured ciphertext blob encrypted under a moto master key in master_keys
    :rtype: bytes
    """
    try:
        key = master_keys[key_id]
    except KeyError:
        is_alias = key_id.startswith("alias/") or ":alias/" in key_id
        id_type = "Alias" if is_alias else "keyId"
        raise NotFoundException(f"{id_type} {key_id} is not found.")

    if plaintext == b"":
        raise ValidationException(
            "1 validation error detected: Value at 'plaintext' failed to satisfy constraint: Member must have length greater than or equal to 1"
        )
    if len(plaintext) > 4096:
        raise ValidationException(
            "1 validation error detected: Value at 'plaintext' failed to satisfy constraint: Member must have length less than or equal to 4096"
        )

    iv = os.urandom(IV_LEN)
    aad = _serialize_encryption_context(encryption_context=encryption_context)

    encryptor = Cipher(
        algorithms.AES(key.key_material), modes.GCM(iv), backend=default_backend()
    ).encryptor()
    encryptor.authenticate_additional_data(aad)
    ciphertext = encryptor.update(plaintext) + encryptor.finalize()
    return _serialize_ciphertext_blob(
        ciphertext=Ciphertext(
            key_id=key_id, iv=iv, ciphertext=ciphertext, tag=encryptor.tag
        )
    )


def decrypt(
    master_keys: Dict[str, Any],
    ciphertext_blob: bytes,
    encryption_context: Dict[str, str],
) -> Tuple[bytes, str]:
    """Decrypt a ciphertext blob using a master key material.

    NOTE: This is not necessarily what KMS does, but it retains the same properties.

    NOTE: This function is NOT compatible with KMS APIs.

    :param dict master_keys: Mapping of a KmsBackend's known master keys
    :param bytes ciphertext_blob: moto-structured ciphertext blob encrypted under a moto master key in master_keys
    :param dict[str, str] encryption_context: KMS-style encryption context
    :returns: plaintext bytes and moto key ID
    :rtype: bytes and str
    """
    try:
        ciphertext = _deserialize_ciphertext_blob(ciphertext_blob=ciphertext_blob)
    except Exception:
        raise InvalidCiphertextException()

    aad = _serialize_encryption_context(encryption_context=encryption_context)

    try:
        key = master_keys[ciphertext.key_id]
    except KeyError:
        raise AccessDeniedException(
            "The ciphertext refers to a customer master key that does not exist, "
            "does not exist in this region, or you are not allowed to access."
        )

    try:
        decryptor = Cipher(
            algorithms.AES(key.key_material),
            modes.GCM(ciphertext.iv, ciphertext.tag),
            backend=default_backend(),
        ).decryptor()
        decryptor.authenticate_additional_data(aad)
        plaintext = decryptor.update(ciphertext.ciphertext) + decryptor.finalize()
    except Exception:
        raise InvalidCiphertextException()

    return plaintext, ciphertext.key_id


def generate_hmac(
    key: bytes,
    message: bytes,
    mac_algorithm: str,
) -> str:
    """
    Returns a base64 encoded HMAC
    """

    algos = {
        "HMAC_SHA_224": hashlib.sha224,
        "HMAC_SHA_256": hashlib.sha256,
        "HMAC_SHA_384": hashlib.sha384,
        "HMAC_SHA_512": hashlib.sha512,
    }

    hmac_val = hmac.new(
        key=key,
        msg=message,
        digestmod=algos[mac_algorithm],
    )

    return base64.b64encode(hmac_val.digest()).decode("utf-8")
